@@ -466,11 +466,14 @@ ntfs_inode *ntfs_pathname_to_inode(ntfs_volume *vol, ntfs_inode *parent,
 			q++;
 		}
 
-		len = ntfs_mbstoucs(p, &unicode, NTFS_MAX_NAME_LEN);
+		len = ntfs_mbstoucs(p, &unicode, MAX_PATH);
 		if (len < 0) {
 			ntfs_log_debug("Couldn't convert name to Unicode: %s.\n", p);
 			err = errno;
 			goto close;
+		} else if (len > NTFS_MAX_NAME_LEN) {
+			err = ENAMETOOLONG;
+			goto close;
 		}
 
 		inum = ntfs_inode_lookup_by_name(ni, unicode, len);
@@ -482,7 +485,10 @@ ntfs_inode *ntfs_pathname_to_inode(ntfs_volume *vol, ntfs_inode *parent,
 		}
 
 		if (ni != parent)
-			ntfs_inode_close(ni);
+			if (ntfs_inode_close(ni)) {
+				err = errno;
+				goto out;
+			}
 
 		inum = MREF(inum);
 		ni = ntfs_inode_open(vol, inum);
@@ -502,7 +508,9 @@ ntfs_inode *ntfs_pathname_to_inode(ntfs_volume *vol, ntfs_inode *parent,
 	ni = NULL;
 close:
 	if (ni && (ni != parent))
-		ntfs_inode_close(ni);
+		if (ntfs_inode_close(ni) && !err)
+			err = errno;
+out:
 	free(ascii);
 	free(unicode);
 	if (err)
@@ -716,7 +724,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 				le16_to_cpu(dir_ni->mrec->sequence_number)),
 				NTFS_DT_DIR);
 		if (rc)
-			goto done;
+			goto err_out;
 		++*pos;
 	}
 	if (*pos == 1) {
@@ -731,7 +739,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 		rc = filldir(dirent, dotdot, 2, FILE_NAME_POSIX, *pos,
 				parent_mref, NTFS_DT_DIR);
 		if (rc)
-			goto done;
+			goto err_out;
 		++*pos;
 	}
 
@@ -744,8 +752,8 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 	/* Find the index root attribute in the mft record. */
 	if (ntfs_attr_lookup(AT_INDEX_ROOT, NTFS_INDEX_I30, 4, CASE_SENSITIVE, 0, NULL,
 			0, ctx)) {
-		ntfs_log_debug("Index root attribute missing in directory inode "
-				"0x%llx.\n", (unsigned long long)dir_ni->mft_no);
+		ntfs_log_perror("Index root attribute missing in directory inode "
+				"0x%llx", (unsigned long long)dir_ni->mft_no);
 		goto dir_err_out;
 	}
 	/* Get to the index root value. */
@@ -756,7 +764,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 	index_block_size = le32_to_cpu(ir->index_block_size);
 	if (index_block_size < NTFS_BLOCK_SIZE ||
 			index_block_size & (index_block_size - 1)) {
-		ntfs_log_debug("Index block size %u is invalid.\n",
+		ntfs_log_error("Index block size %u is invalid.\n",
 				(unsigned)index_block_size);
 		goto dir_err_out;
 	}
@@ -808,7 +816,7 @@ int ntfs_readdir(ntfs_inode *dir_ni, s64 *pos,
 		if (rc) {
 			ntfs_attr_put_search_ctx(ctx);
 			ctx = NULL;
-			goto done;
+			goto err_out;
 		}
 	}
 	ntfs_attr_put_search_ctx(ctx);
@@ -842,7 +850,7 @@ skip_index_root:
 
 	bmp_pos = ia_pos >> index_block_size_bits;
 	if (bmp_pos >> 3 >= bmp_na->data_size) {
-		ntfs_log_debug("Current index position exceeds index bitmap "
+		ntfs_log_error("Current index position exceeds index bitmap "
 				"size.\n");
 		goto dir_err_out;
 	}
@@ -899,7 +907,7 @@ find_next_index_buffer:
 	ia_start = ia_pos & ~(s64)(index_block_size - 1);
 	if (sle64_to_cpu(ia->index_block_vcn) != ia_start >>
 			index_vcn_size_bits) {
-		ntfs_log_debug("Actual VCN (0x%llx) of index buffer is different "
+		ntfs_log_error("Actual VCN (0x%llx) of index buffer is different "
 				"from expected VCN (0x%llx) in inode 0x%llx.\n",
 				(long long)sle64_to_cpu(ia->index_block_vcn),
 				(long long)ia_start >> index_vcn_size_bits,
@@ -907,7 +915,7 @@ find_next_index_buffer:
 		goto dir_err_out;
 	}
 	if (le32_to_cpu(ia->index.allocated_size) + 0x18 != index_block_size) {
-		ntfs_log_debug("Index buffer (VCN 0x%llx) of directory inode 0x%llx "
+		ntfs_log_error("Index buffer (VCN 0x%llx) of directory inode 0x%llx "
 				"has a size (%u) differing from the directory "
 				"specified size (%u).\n", (long long)ia_start >>
 				index_vcn_size_bits,
@@ -918,7 +926,7 @@ find_next_index_buffer:
 	}
 	index_end = (u8*)&ia->index + le32_to_cpu(ia->index.index_length);
 	if (index_end > (u8*)ia + index_block_size) {
-		ntfs_log_debug("Size of index buffer (VCN 0x%llx) of directory inode "
+		ntfs_log_error("Size of index buffer (VCN 0x%llx) of directory inode "
 				"0x%llx exceeds maximum size.\n",
 				(long long)ia_start >> index_vcn_size_bits,
 				(unsigned long long)dir_ni->mft_no);
@@ -940,7 +948,7 @@ find_next_index_buffer:
 				sizeof(INDEX_ENTRY_HEADER) > index_end ||
 				(u8*)ie + le16_to_cpu(ie->key_length) >
 				index_end) {
-			ntfs_log_debug("Index entry out of bounds in directory inode "
+			ntfs_log_error("Index entry out of bounds in directory inode "
 				"0x%llx.\n", (unsigned long long)dir_ni->mft_no);
 			goto dir_err_out;
 		}
@@ -957,7 +965,7 @@ find_next_index_buffer:
 		rc = ntfs_filldir(dir_ni, pos, index_vcn_size_bits,
 				INDEX_TYPE_ALLOCATION, ia, ie, dirent, filldir);
 		if (rc)
-			goto done;
+			goto err_out;
 	}
 	goto find_next_index_buffer;
 EOD:
@@ -970,13 +978,7 @@ done:
 		ntfs_attr_close(bmp_na);
 	if (ia_na)
 		ntfs_attr_close(ia_na);
-#ifdef DEBUG
-	if (!rc)
-		ntfs_log_debug("EOD, *pos 0x%llx, returning 0.\n", (long long)*pos);
-	else
-		ntfs_log_debug("filldir returned %i, *pos 0x%llx, returning 0.\n",
-				rc, (long long)*pos);
-#endif
+	ntfs_log_debug("EOD, *pos 0x%llx, returning 0.\n", (long long)*pos);
 	return 0;
 dir_err_out:
 	errno = EIO;
@@ -1045,12 +1047,15 @@ static ntfs_inode *__ntfs_create(ntfs_inode *dir_ni,
 		errno = EINVAL;
 		return NULL;
 	}
-	/* Allocate MFT record for new file. */
-	ni = ntfs_mft_record_alloc(dir_ni->vol, NULL);
-	if (!ni) {
-		ntfs_log_perror("Could not allocate new MFT record");
+	
+	if (dir_ni->flags & FILE_ATTR_REPARSE_POINT) {
+		errno = EOPNOTSUPP;
 		return NULL;
 	}
+	
+	ni = ntfs_mft_record_alloc(dir_ni->vol, NULL);
+	if (!ni)
+		return NULL;
 	/*
 	 * Create STANDARD_INFORMATION attribute. Write STANDARD_INFORMATION
 	 * version 1.2, windows will upgrade it to version 3 if needed.
@@ -1525,8 +1530,8 @@ out:
 		ntfs_attr_put_search_ctx(actx);
 	if (ictx)
 		ntfs_index_ctx_put(ictx);
-	if (ni)
-		ntfs_inode_close(ni);
+	if (ni && ntfs_inode_close(ni) && !err)
+		err = errno;
 	if (err) {
 		errno = err;
 		ntfs_log_debug("Could not delete file: %s\n", strerror(errno));
@@ -1567,6 +1572,12 @@ int ntfs_link(ntfs_inode *ni, ntfs_inode *dir_ni, ntfschar *name, u8 name_len)
 		ntfs_log_perror("ntfs_link wrong arguments");
 		goto err_out;
 	}
+	
+	if (ni->flags & FILE_ATTR_REPARSE_POINT) {
+		err = EOPNOTSUPP;
+		goto err_out;
+	}
+	
 	/* Create FILE_NAME attribute. */
 	fn_len = sizeof(FILE_NAME_ATTR) + name_len * sizeof(ntfschar);
 	fn = ntfs_calloc(fn_len);
@@ -1628,7 +1639,6 @@ rollback_failed:
 err_out:
 	free(fn);
 	errno = err;
-	ntfs_log_perror("Hard link failed");
 	return -1;
 }
 
