@@ -51,6 +51,13 @@
 #include "logging.h"
 #include "misc.h"
 
+ntfs_inode *ntfs_inode_base(ntfs_inode *ni)
+{
+	if (ni->nr_extents == -1)
+		return ni->base_ni;
+	return ni;
+}
+
 /**
  * ntfs_inode_mark_dirty - set the inode (and its base inode if it exists) dirty
  * @ni:		ntfs inode to set dirty
@@ -107,15 +114,16 @@ ntfs_inode *ntfs_inode_allocate(ntfs_volume *vol)
  *
  * Returns:
  */
-static int __ntfs_inode_release(ntfs_inode *ni)
+static void __ntfs_inode_release(ntfs_inode *ni)
 {
 	if (NInoDirty(ni))
-		ntfs_log_debug("Eeek. Discarding dirty inode!\n");
+		ntfs_log_error("Releasing dirty inode %lld!\n", 
+			       (long long)ni->mft_no);
 	if (NInoAttrList(ni) && ni->attr_list)
 		free(ni->attr_list);
 	free(ni->mrec);
 	free(ni);
-	return 0;
+	return;
 }
 
 /**
@@ -271,7 +279,7 @@ int ntfs_inode_close(ntfs_inode *ni)
 	if (!ni)
 		return 0;
 
-	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long) ni->mft_no);
+	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long)ni->mft_no);
 
 	/* If we have dirty metadata, write it out. */
 	if (NInoDirty(ni) || NInoAttrListDirty(ni)) {
@@ -330,11 +338,18 @@ int ntfs_inode_close(ntfs_inode *ni)
 			i = -1;
 			break;
 		}
+		
+		/* 
+		 *  We could successfully sync, so only log this error
+		 *  and try to sync other inode extents too.
+		 */
 		if (i != -1)
-			ntfs_log_debug("Extent inode was not attached to base inode! "
-					"Weird! Continuing regardless.\n");
+			ntfs_log_error("Extent inode %lld was not found\n",
+				       (long long)ni->mft_no);
 	}
-	return __ntfs_inode_release(ni);
+	
+	__ntfs_inode_release(ni);
+	return 0;
 }
 
 /**
@@ -371,11 +386,14 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 
 	if (!base_ni) {
 		errno = EINVAL;
+		ntfs_log_perror("%s", __FUNCTION__);
 		return NULL;
 	}
+	
 	ntfs_log_trace("Opening extent inode 0x%llx (base mft record 0x%llx).\n",
 			(unsigned long long)mft_no,
 			(unsigned long long)base_ni->mft_no);
+	
 	/* Is the extent inode already open and attached to the base inode? */
 	if (base_ni->nr_extents > 0) {
 		extent_nis = base_ni->extent_nis;
@@ -389,12 +407,11 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 			seq_no = MSEQNO_LE(mref);
 			if (seq_no && seq_no != le16_to_cpu(
 					ni->mrec->sequence_number)) {
-				ntfs_log_debug("Found stale extent mft reference! "
-					"Corrupt file system. Run chkdsk.\n");
 				errno = EIO;
+				ntfs_log_perror("Found stale extent mft "
+					"reference mft=%lld", ni->mft_no);
 				return NULL;
 			}
-			/* We are done, return the extent inode. */
 			return ni;
 		}
 	}
@@ -403,8 +420,10 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 	if (!ni)
 		return NULL;
 	if (ntfs_file_record_read(base_ni->vol, le64_to_cpu(mref), &ni->mrec,
-			NULL))
+			NULL)) {
+		ntfs_log_perror("ntfs_file_record_read failed #2");
 		goto err_out;
+	}
 	ni->mft_no = mft_no;
 	ni->nr_extents = -1;
 	ni->base_ni = base_ni;
@@ -425,10 +444,7 @@ ntfs_inode *ntfs_extent_inode_open(ntfs_inode *base_ni, const MFT_REF mref)
 	base_ni->extent_nis[base_ni->nr_extents++] = ni;
 	return ni;
 err_out:
-	i = errno;
 	__ntfs_inode_release(ni);
-	errno = i;
-	ntfs_log_perror("Failed to open extent inode");
 	return NULL;
 }
 
@@ -536,12 +552,11 @@ static int ntfs_inode_sync_file_name(ntfs_inode *ni)
 	FILE_NAME_ATTR *fn;
 	int err = 0;
 
-	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long) ni->mft_no);
+	ntfs_log_trace("Entering for inode %lld\n", (long long)ni->mft_no);
 
 	ctx = ntfs_attr_get_search_ctx(ni, NULL);
 	if (!ctx) {
 		err = errno;
-		ntfs_log_trace("Failed to get attribute search context.\n");
 		goto err_out;
 	}
 	/* Walk through all FILE_NAME attributes and update them. */
@@ -550,27 +565,29 @@ static int ntfs_inode_sync_file_name(ntfs_inode *ni)
 				le16_to_cpu(ctx->attr->value_offset));
 		if (MREF_LE(fn->parent_directory) == ni->mft_no) {
 			/*
-			 * WARNING: We cheater here and obtain 2 attribute
+			 * WARNING: We cheat here and obtain 2 attribute
 			 * search contexts for one inode (first we obtained
 			 * above, second will be obtained inside
 			 * ntfs_index_lookup), it's acceptable for library,
-			 * but will lock kernel.
+			 * but will deadlock in the kernel.
 			 */
 			index_ni = ni;
 		} else
-			index_ni = ntfs_inode_open(ni->vol,
-				le64_to_cpu(fn->parent_directory));
+			index_ni = ntfs_inode_open(ni->vol, 
+					le64_to_cpu(fn->parent_directory));
 		if (!index_ni) {
 			if (!err)
 				err = errno;
-			ntfs_log_trace("Failed to open inode with index.\n");
+			ntfs_log_perror("Failed to open inode %lld with index",
+					le64_to_cpu(fn->parent_directory));
 			continue;
 		}
 		ictx = ntfs_index_ctx_get(index_ni, NTFS_INDEX_I30, 4);
 		if (!ictx) {
 			if (!err)
 				err = errno;
-			ntfs_log_trace("Failed to get index context.\n");
+			ntfs_log_perror("Failed to get index ctx, inode %lld",
+					(long long)index_ni->mft_no);
 			ntfs_inode_close(index_ni);
 			continue;
 		}
@@ -581,7 +598,8 @@ static int ntfs_inode_sync_file_name(ntfs_inode *ni)
 				else
 					err = errno;
 			}
-			ntfs_log_trace("Index lookup failed.\n");
+			ntfs_log_perror("Index lookup failed, inode %lld",
+					(long long)index_ni->mft_no);
 			ntfs_index_ctx_put(ictx);
 			ntfs_inode_close(index_ni);
 			continue;
@@ -599,13 +617,14 @@ static int ntfs_inode_sync_file_name(ntfs_inode *ni)
 		fn->last_access_time = utc2ntfs(ni->last_access_time);
 		ntfs_index_entry_mark_dirty(ictx);
 		ntfs_index_ctx_put(ictx);
-		if (ni != index_ni)
-			ntfs_inode_close(index_ni);
+		if ((ni != index_ni) && ntfs_inode_close(index_ni) && !err)
+			err = errno;
 	}
 	/* Check for real error occurred. */
 	if (errno != ENOENT) {
 		err = errno;
-		ntfs_log_trace("Attribute lookup failed.\n");
+		ntfs_log_perror("Attribute lookup failed, inode %lld",
+				(long long)ni->mft_no);
 		goto err_out;
 	}
 	ntfs_attr_put_search_ctx(ctx);
@@ -646,10 +665,11 @@ int ntfs_inode_sync(ntfs_inode *ni)
 
 	if (!ni) {
 		errno = EINVAL;
+		ntfs_log_error("Failed to sync NULL inode\n");
 		return -1;
 	}
 
-	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long) ni->mft_no);
+	ntfs_log_trace("Entering for inode %lld\n", (long long)ni->mft_no);
 
 	/* Update STANDARD_INFORMATION. */
 	if ((ni->mrec->flags & MFT_RECORD_IN_USE) && ni->nr_extents != -1 &&
@@ -659,19 +679,22 @@ int ntfs_inode_sync(ntfs_inode *ni)
 			if (err != EIO)
 				err = EBUSY;
 		}
-		ntfs_log_trace("Failed to sync standard information.\n");
+		ntfs_log_perror("Failed to sync standard info (inode %lld)",
+				(long long)ni->mft_no);
 	}
 
 	/* Update FILE_NAME's in the index. */
 	if ((ni->mrec->flags & MFT_RECORD_IN_USE) && ni->nr_extents != -1 &&
 			NInoFileNameTestAndClearDirty(ni) &&
+			NInoParentMtimeUpdate(ni) &&
 			ntfs_inode_sync_file_name(ni)) {
 		if (!err || errno == EIO) {
 			err = errno;
 			if (err != EIO)
 				err = EBUSY;
 		}
-		ntfs_log_trace("Failed to sync FILE_NAME attributes.\n");
+		ntfs_log_perror("Failed to sync FILE_NAME (inode %lld)",
+				(long long)ni->mft_no);
 		NInoFileNameSetDirty(ni);
 	}
 
@@ -686,33 +709,37 @@ int ntfs_inode_sync(ntfs_inode *ni)
 				err = errno;
 				if (err != EIO)
 					err = EBUSY;
-				ntfs_log_trace("Attribute list sync failed (open "
-						"failed).\n");
+				ntfs_log_perror("Attribute list sync failed "
+						"(open, inode %lld)",
+						(long long)ni->mft_no);
 			}
 			NInoAttrListSetDirty(ni);
-		} else {
-			if (na->data_size == ni->attr_list_size) {
-				if (ntfs_attr_pwrite(na, 0, ni->attr_list_size,
-							ni->attr_list) !=
-							ni->attr_list_size) {
-					if (!err || errno == EIO) {
-						err = errno;
-						if (err != EIO)
-							err = EBUSY;
-						ntfs_log_trace("Attribute list sync "
-							"failed (write failed).\n");
-					}
-					NInoAttrListSetDirty(ni);
+			goto sync_inode;
+		} 
+		
+		if (na->data_size == ni->attr_list_size) {
+			if (ntfs_attr_pwrite(na, 0, ni->attr_list_size,
+				        ni->attr_list) != ni->attr_list_size) {
+				if (!err || errno == EIO) {
+					err = errno;
+					if (err != EIO)
+						err = EBUSY;
+					ntfs_log_perror("Attribute list sync "
+						"failed (write, inode %lld)",
+						(long long)ni->mft_no);
 				}
-			} else {
-				err = EIO;
-				ntfs_log_trace("Attribute list sync failed (invalid size).\n");
 				NInoAttrListSetDirty(ni);
 			}
-			ntfs_attr_close(na);
+		} else {
+			err = EIO;
+			ntfs_log_error("Attribute list sync failed (bad size, "
+				       "inode %lld)\n", (long long)ni->mft_no);
+			NInoAttrListSetDirty(ni);
 		}
+		ntfs_attr_close(na);
 	}
-
+	
+sync_inode:
 	/* Write this inode out to the $MFT (and $MFTMirr if applicable). */
 	if (NInoTestAndClearDirty(ni)) {
 		if (ntfs_mft_record_write(ni->vol, ni->mft_no, ni->mrec)) {
@@ -722,7 +749,8 @@ int ntfs_inode_sync(ntfs_inode *ni)
 					err = EBUSY;
 			}
 			NInoSetDirty(ni);
-			ntfs_log_trace("Base MFT record sync failed.\n");
+			ntfs_log_perror("MFT record sync failed, inode %lld",
+					(long long)ni->mft_no);
 		}
 	}
 
@@ -734,18 +762,21 @@ int ntfs_inode_sync(ntfs_inode *ni)
 			ntfs_inode *eni;
 
 			eni = ni->extent_nis[i];
-			if (NInoTestAndClearDirty(eni)) {
-				if (ntfs_mft_record_write(eni->vol, eni->mft_no,
-						eni->mrec)) {
-					if (!err || errno == EIO) {
-						err = errno;
-						if (err != EIO)
-							err = EBUSY;
-					}
-					NInoSetDirty(eni);
-					ntfs_log_trace("Extent MFT record sync "
-							"failed.\n");
+			if (!NInoTestAndClearDirty(eni))
+				continue;
+			
+			if (ntfs_mft_record_write(eni->vol, eni->mft_no, 
+						  eni->mrec)) {
+				if (!err || errno == EIO) {
+					err = errno;
+					if (err != EIO)
+						err = EBUSY;
 				}
+				NInoSetDirty(eni);
+				ntfs_log_perror("Extent MFT record sync failed,"
+						" inode %lld/%lld",
+						(long long)ni->mft_no,
+						(long long)eni->mft_no);
 			}
 		}
 	}
@@ -777,16 +808,16 @@ int ntfs_inode_add_attrlist(ntfs_inode *ni)
 	ntfs_attr *na;
 
 	if (!ni) {
-		ntfs_log_trace("Invalid arguments.\n");
 		errno = EINVAL;
+		ntfs_log_perror("%s", __FUNCTION__);
 		return -1;
 	}
 
-	ntfs_log_trace("Entering for inode 0x%llx.\n", (long long) ni->mft_no);
+	ntfs_log_trace("inode %llu\n", (unsigned long long) ni->mft_no);
 
 	if (NInoAttrList(ni) || ni->nr_extents) {
-		ntfs_log_trace("Inode already has got attribute list.\n");
 		errno = EEXIST;
+		ntfs_log_perror("Inode already has attribute list");
 		return -1;
 	}
 
@@ -794,7 +825,6 @@ int ntfs_inode_add_attrlist(ntfs_inode *ni)
 	ctx = ntfs_attr_get_search_ctx(ni, NULL);
 	if (!ctx) {
 		err = errno;
-		ntfs_log_trace("Couldn't get search context.\n");
 		goto err_out;
 	}
 	/* Walk through all attributes. */
@@ -804,7 +834,7 @@ int ntfs_inode_add_attrlist(ntfs_inode *ni)
 		
 		if (ctx->attr->type == AT_ATTRIBUTE_LIST) {
 			err = EIO;
-			ntfs_log_trace("Eeek! Attribute list already present.\n");
+			ntfs_log_perror("Attribute list already present");
 			goto put_err_out;
 		}
 		
@@ -844,7 +874,8 @@ int ntfs_inode_add_attrlist(ntfs_inode *ni)
 	/* Check for real error occurred. */
 	if (errno != ENOENT) {
 		err = errno;
-		ntfs_log_trace("Attribute lookup failed.\n");
+		ntfs_log_perror("%s: Attribute lookup failed, inode %lld",
+				__FUNCTION__, (long long)ni->mft_no);
 		goto put_err_out;
 	}
 
@@ -862,8 +893,7 @@ int ntfs_inode_add_attrlist(ntfs_inode *ni)
 				offsetof(ATTR_RECORD, resident_end))) {
 			/* Failed to free space. */
 			err = errno;
-			ntfs_log_trace("Failed to free space for "
-					"$ATTRIBUTE_LIST.\n");
+			ntfs_log_perror("Failed to free space for attrlist");
 			goto rollback;
 		}
 	}
@@ -872,7 +902,7 @@ int ntfs_inode_add_attrlist(ntfs_inode *ni)
 	if (ntfs_resident_attr_record_add(ni,
 				AT_ATTRIBUTE_LIST, NULL, 0, NULL, 0, 0) < 0) {
 		err = errno;
-		ntfs_log_trace("Couldn't add $ATTRIBUTE_LIST to MFT record.\n");
+		ntfs_log_perror("Couldn't add $ATTRIBUTE_LIST to MFT");
 		goto rollback;
 	}
 
@@ -880,12 +910,12 @@ int ntfs_inode_add_attrlist(ntfs_inode *ni)
 	na = ntfs_attr_open(ni, AT_ATTRIBUTE_LIST, AT_UNNAMED, 0);
 	if (!na) {
 		err = errno;
-		ntfs_log_trace("Failed to open just added $ATTRIBUTE_LIST.\n");
+		ntfs_log_perror("Failed to open just added $ATTRIBUTE_LIST");
 		goto remove_attrlist_record;
 	}
 	if (ntfs_attr_truncate(na, al_len)) {
 		err = errno;
-		ntfs_log_trace("Failed to resize just added $ATTRIBUTE_LIST.\n");
+		ntfs_log_perror("Failed to resize just added $ATTRIBUTE_LIST");
 		ntfs_attr_close(na);
 		goto remove_attrlist_record;;
 	}
@@ -903,11 +933,9 @@ remove_attrlist_record:
 	if (!ntfs_attr_lookup(AT_ATTRIBUTE_LIST, NULL, 0,
 				CASE_SENSITIVE, 0, NULL, 0, ctx)) {
 		if (ntfs_attr_record_rm(ctx))
-			ntfs_log_trace("Rollback failed. Failed to remove attribute "
-					"list record.\n");
+			ntfs_log_perror("Rollback failed to remove attrlist");
 	} else
-		ntfs_log_trace("Rollback failed. Couldn't find attribute list "
-				"record.\n");
+		ntfs_log_perror("Rollback failed to find attrlist");
 	/* Setup back in-memory runlist. */
 	ni->attr_list = al;
 	ni->attr_list_size = al_len;
@@ -927,11 +955,10 @@ rollback:
 						sle64_to_cpu(ale->lowest_vcn),
 						NULL, 0, ctx)) {
 				if (ntfs_attr_record_move_to(ctx, ni))
-					ntfs_log_trace("Rollback failed. Couldn't "
-						"back attribute to base MFT record.\n");
+					ntfs_log_perror("Rollback failed to "
+							"move attribute");
 			} else
-				ntfs_log_trace("Rollback failed. ntfs_attr_lookup "
-						"failed.\n");
+				ntfs_log_perror("Rollback failed to find attr");
 			ntfs_attr_reinit_search_ctx(ctx);
 		}
 		ale = (ATTR_LIST_ENTRY*)((u8*)ale + le16_to_cpu(ale->length));
@@ -962,13 +989,13 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 	int freed, err;
 
 	if (!ni || size < 0) {
-		ntfs_log_trace("Invalid arguments.\n");
 		errno = EINVAL;
+		ntfs_log_perror("%s: ni=%p size=%d", __FUNCTION__, ni, size);
 		return -1;
 	}
 
-	ntfs_log_trace("Entering for inode 0x%llx, size %d.\n",
-			(long long) ni->mft_no, size);
+	ntfs_log_trace("Entering for inode %lld, size %d\n",
+		       (unsigned long long)ni->mft_no, size);
 
 	freed = (le32_to_cpu(ni->mrec->bytes_allocated) -
 				le32_to_cpu(ni->mrec->bytes_in_use));
@@ -977,12 +1004,8 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 		return 0;
 
 	ctx = ntfs_attr_get_search_ctx(ni, NULL);
-	if (!ctx) {
-		err = errno;
-		ntfs_log_trace("Failed to get attribute search context.\n");
-		errno = err;
+	if (!ctx)
 		return -1;
-	}
 
 	/*
 	 * Chkdsk complain if $STANDARD_INFORMATION is not in the base MFT
@@ -1002,7 +1025,7 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 				0, ctx)) {
 		if (errno != ENOENT) {
 			err = errno;
-			ntfs_log_trace("Attribute lookup failed.\n");
+			ntfs_log_perror("%s: attr lookup failed #2", __FUNCTION__);
 			goto put_err_out;
 		}
 		if (ctx->attr->type == AT_END) {
@@ -1023,7 +1046,7 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 						0, NULL, 0, ctx)) {
 				err = errno;
 				if (errno != ENOENT) {
-					ntfs_log_trace("Attribute lookup failed.\n");
+					ntfs_log_perror("Attr lookup failed #2");
 				} else
 					err = ENOSPC;
 				goto put_err_out;
@@ -1035,7 +1058,7 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 		/* Move away attribute. */
 		if (ntfs_attr_record_move_away(ctx, 0)) {
 			err = errno;
-			ntfs_log_trace("Failed to move out attribute.\n");
+			ntfs_log_perror("Failed to move out attribute #2");
 			break;
 		}
 		freed += record_size;
@@ -1055,7 +1078,7 @@ int ntfs_inode_free_space(ntfs_inode *ni, int size)
 				NULL, 0, ctx)) {
 			if (errno != ENOENT) {
 				err = errno;
-				ntfs_log_trace("Attribute lookup failed.\n");
+				ntfs_log_perror("Attr lookup #2 failed");
 				break;
 			}
 			if (ctx->attr->type == AT_END) {
@@ -1099,7 +1122,7 @@ void ntfs_inode_update_atime(ntfs_inode *ni)
  */
 void ntfs_inode_update_time(ntfs_inode *ni)
 {
-	if (!NVolReadOnly(ni->vol) && 
+	if (!NVolReadOnly(ni->vol) && NInoMtimeUpdate(ni) &&
 	    (ni->mft_no >= FILE_first_user || ni->mft_no == FILE_root)) {
 		time_t now;
 
