@@ -81,9 +81,12 @@ int ntfs_mft_records_read(const ntfs_volume *vol, const MFT_REF mref,
 	s64 br;
 	VCN m;
 
-	ntfs_log_trace("Entering for inode 0x%llx.\n", MREF(mref));
+	ntfs_log_trace("Entering for inode %lld\n", MREF(mref));
+	
 	if (!vol || !vol->mft_na || !b || count < 0) {
 		errno = EINVAL;
+		ntfs_log_perror("%s: b=%p  count=%lld  mft=%lld", __FUNCTION__,
+				b, count, MREF(mref));
 		return -1;
 	}
 	m = MREF(mref);
@@ -246,6 +249,7 @@ int ntfs_file_record_read(const ntfs_volume *vol, const MFT_REF mref,
 
 	if (!vol || !mrec) {
 		errno = EINVAL;
+		ntfs_log_perror("%s: mrec=%p", __FUNCTION__, mrec);
 		return -1;
 	}
 	m = *mrec;
@@ -256,23 +260,28 @@ int ntfs_file_record_read(const ntfs_volume *vol, const MFT_REF mref,
 	}
 	if (ntfs_mft_record_read(vol, mref, m)) {
 		err = errno;
-		goto read_failed;
+		ntfs_log_perror("ntfs_mft_record_read failed");
+		goto err_out;
+	}
+	err = EIO;
+	if (!ntfs_is_file_record(m->magic)) {
+		ntfs_log_perror("Record %llu has no FILE magic", MREF(mref));
+		goto err_out;
+	}
+	if (MSEQNO(mref) && MSEQNO(mref) != le16_to_cpu(m->sequence_number)) {
+		ntfs_log_perror("Record %llu has wrong SeqNo", MREF(mref));
+		goto err_out;
 	}
-	if (!ntfs_is_file_record(m->magic))
-		goto file_corrupt;
-	if (MSEQNO(mref) && MSEQNO(mref) != le16_to_cpu(m->sequence_number))
-		goto file_corrupt;
 	a = (ATTR_RECORD*)((char*)m + le16_to_cpu(m->attrs_offset));
-	if (p2n(a) < p2n(m) || (char*)a > (char*)m + vol->mft_record_size)
-		goto file_corrupt;
+	if (p2n(a) < p2n(m) || (char*)a > (char*)m + vol->mft_record_size) {
+		ntfs_log_perror("Record %llu is corrupt", MREF(mref));
+		goto err_out;
+	}
 	*mrec = m;
 	if (attr)
 		*attr = a;
 	return 0;
-file_corrupt:
-	ntfs_log_debug("ntfs_file_record_read(): file is corrupt.\n");
-	err = EIO;
-read_failed:
+err_out:
 	if (m != *mrec)
 		free(m);
 	errno = err;
@@ -299,6 +308,7 @@ int ntfs_mft_record_layout(const ntfs_volume *vol, const MFT_REF mref,
 
 	if (!vol || !mrec) {
 		errno = EINVAL;
+		ntfs_log_perror("%s: mrec=%p", __FUNCTION__, mrec);
 		return -1;
 	}
 	/* Aligned to 2-byte boundary. */
@@ -307,8 +317,8 @@ int ntfs_mft_record_layout(const ntfs_volume *vol, const MFT_REF mref,
 	else {
 		/* Abort if mref is > 32 bits. */
 		if (MREF(mref) & 0x0000ffff00000000ull) {
-			ntfs_log_debug("Mft reference exceeds 32 bits!\n");
 			errno = ERANGE;
+			ntfs_log_perror("Mft reference exceeds 32 bits");
 			return -1;
 		}
 		mrec->usa_ofs = cpu_to_le16((sizeof(MFT_RECORD) + 1) & ~1);
@@ -371,29 +381,22 @@ int ntfs_mft_record_layout(const ntfs_volume *vol, const MFT_REF mref,
 int ntfs_mft_record_format(const ntfs_volume *vol, const MFT_REF mref)
 {
 	MFT_RECORD *m;
-	int err;
+	int ret = -1;
 
-	if (!vol || !vol->mft_na) {
-		errno = EINVAL;
-		return -1;
-	}
 	m = ntfs_calloc(vol->mft_record_size);
 	if (!m)
 		return -1;
-	if (ntfs_mft_record_layout(vol, mref, m)) {
-		err = errno;
-		free(m);
-		errno = err;
-		return -1;
-	}
-	if (ntfs_mft_record_write(vol, mref, m)) {
-		err = errno;
-		free(m);
-		errno = err;
-		return -1;
-	}
+	
+	if (ntfs_mft_record_layout(vol, mref, m))
+		goto free_m;
+	
+	if (ntfs_mft_record_write(vol, mref, m))
+		goto free_m;
+	
+	ret = 0;
+free_m:
 	free(m);
-	return 0;
+	return ret;
 }
 
 static const char *es = "  Leaving inconsistent metadata.  Run chkdsk.";
@@ -488,8 +491,7 @@ static int ntfs_mft_bitmap_find_free_rec(ntfs_volume *vol, ntfs_inode *base_ni)
 			size = ll;
 		ll = ntfs_attr_pread(mftbmp_na, ofs, size, buf);
 		if (ll < 0) {
-			ntfs_log_error("Failed to read mft bitmap "
-					"attribute, aborting.\n");
+			ntfs_log_perror("Failed to read $MFT bitmap");
 			free(buf);
 			return -1;
 		}
@@ -566,22 +568,16 @@ static int ntfs_mft_bitmap_extend_allocation(ntfs_volume *vol)
 {
 	LCN lcn;
 	s64 ll = 0; /* silence compiler warning */
-	ntfs_attr *mftbmp_na, *lcnbmp_na;
+	ntfs_attr *mftbmp_na;
 	runlist_element *rl, *rl2 = NULL; /* silence compiler warning */
 	ntfs_attr_search_ctx *ctx;
 	MFT_RECORD *m = NULL; /* silence compiler warning */
 	ATTR_RECORD *a = NULL; /* silence compiler warning */
-	int ret, mp_size;
+	int err, mp_size;
 	u32 old_alen = 0; /* silence compiler warning */
-	u8 b, tb;
-	struct {
-		u8 added_cluster:1;
-		u8 added_run:1;
-		u8 mp_rebuilt:1;
-	} status = { 0, 0, 0 };
+	BOOL mp_rebuilt = FALSE;
 
 	mftbmp_na = vol->mftbmp_na;
-	lcnbmp_na = vol->lcnbmp_na;
 	/*
 	 * Determine the last lcn of the mft bitmap.  The allocated size of the
 	 * mft bitmap cannot be zero so we are ok to do this.
@@ -596,70 +592,38 @@ static int ntfs_mft_bitmap_extend_allocation(ntfs_volume *vol)
 		return -1;
 	}
 	lcn = rl->lcn + rl->length;
-	/*
-	 * Attempt to get the cluster following the last allocated cluster by
-	 * hand as it may be in the MFT zone so the allocator would not give it
-	 * to us.
-	 */
-	ret = (int)ntfs_attr_pread(lcnbmp_na, lcn >> 3, 1, &b);
-	if (ret < 0) {
-		ntfs_log_error("Failed to read from lcn bitmap.\n");
+	
+	rl2 = ntfs_cluster_alloc(vol, rl[1].vcn, 1, lcn, DATA_ZONE);
+	if (!rl2) {
+		ntfs_log_error("Failed to allocate a cluster for "
+				"the mft bitmap.\n");
 		return -1;
 	}
-	ntfs_log_debug("Read %i byte%s.\n", ret, ret == 1 ? "" : "s");
-	tb = 1 << (lcn & 7ull);
-	if (ret == 1 && b != 0xff && !(b & tb)) {
-		/* Next cluster is free, allocate it. */
-		b |= tb;
-		ret = (int)ntfs_attr_pwrite(lcnbmp_na, lcn >> 3, 1, &b);
-		if (ret < 1) {
-			ntfs_log_error("Failed to write to lcn "
-					"bitmap.\n");
-			if (!ret)
-				errno = EIO;
-			return -1;
-		}
-		/* Update the mft bitmap runlist. */
-		rl->length++;
-		rl[1].vcn++;
-		status.added_cluster = 1;
-		ntfs_log_debug("Appending one cluster to mft bitmap.\n");
-	} else {
-		/* Allocate a cluster from the DATA_ZONE. */
-		rl2 = ntfs_cluster_alloc(vol, rl[1].vcn, 1, lcn, DATA_ZONE);
-		if (!rl2) {
-			ntfs_log_error("Failed to allocate a cluster for "
-					"the mft bitmap.\n");
-			return -1;
-		}
-		rl = ntfs_runlists_merge(mftbmp_na->rl, rl2);
-		if (!rl) {
-			ret = errno;
-			ntfs_log_error("Failed to merge runlists for mft "
-					"bitmap.\n");
-			if (ntfs_cluster_free_from_rl(vol, rl2))
-				ntfs_log_error("Failed to deallocate "
-						"cluster.%s\n", es);
-			free(rl2);
-			errno = ret;
-			return -1;
-		}
-		mftbmp_na->rl = rl;
-		status.added_run = 1;
-		ntfs_log_debug("Adding one run to mft bitmap.\n");
-		/* Find the last run in the new runlist. */
-		for (; rl[1].length; rl++)
-			;
+	rl = ntfs_runlists_merge(mftbmp_na->rl, rl2);
+	if (!rl) {
+		err = errno;
+		ntfs_log_error("Failed to merge runlists for mft "
+				"bitmap.\n");
+		if (ntfs_cluster_free_from_rl(vol, rl2))
+			ntfs_log_error("Failed to deallocate "
+					"cluster.%s\n", es);
+		free(rl2);
+		errno = err;
+		return -1;
 	}
+	mftbmp_na->rl = rl;
+	ntfs_log_debug("Adding one run to mft bitmap.\n");
+	/* Find the last run in the new runlist. */
+	for (; rl[1].length; rl++)
+		;
 	/*
 	 * Update the attribute record as well.  Note: @rl is the last
 	 * (non-terminator) runlist element of mft bitmap.
 	 */
 	ctx = ntfs_attr_get_search_ctx(mftbmp_na->ni, NULL);
-	if (!ctx) {
-		ntfs_log_error("Failed to get search context.\n");
+	if (!ctx)
 		goto undo_alloc;
-	}
+
 	if (ntfs_attr_lookup(mftbmp_na->type, mftbmp_na->name,
 			mftbmp_na->name_len, 0, rl[1].vcn, NULL, 0, ctx)) {
 		ntfs_log_error("Failed to find last attribute extent of "
@@ -696,7 +660,7 @@ static int ntfs_mft_bitmap_extend_allocation(ntfs_volume *vol)
 		errno = EOPNOTSUPP;
 		goto undo_alloc;
 	}
-	status.mp_rebuilt = 1;
+	mp_rebuilt = TRUE;
 	/* Generate the mapping pairs array directly into the attr record. */
 	if (ntfs_mapping_pairs_build(vol, (u8*)a +
 			le16_to_cpu(a->mapping_pairs_offset), mp_size, rl2, ll,
@@ -734,7 +698,7 @@ static int ntfs_mft_bitmap_extend_allocation(ntfs_volume *vol)
 	ntfs_attr_put_search_ctx(ctx);
 	return 0;
 restore_undo_alloc:
-	ret = errno;
+	err = errno;
 	ntfs_attr_reinit_search_ctx(ctx);
 	if (ntfs_attr_lookup(mftbmp_na->type, mftbmp_na->name,
 			mftbmp_na->name_len, 0, rl[1].vcn, NULL, 0, ctx)) {
@@ -746,29 +710,27 @@ restore_undo_alloc:
 		 * The only thing that is now wrong is ->allocated_size of the
 		 * base attribute extent which chkdsk should be able to fix.
 		 */
-		errno = ret;
+		errno = err;
 		return -1;
 	}
 	m = ctx->mrec;
 	a = ctx->attr;
 	a->highest_vcn = cpu_to_sle64(rl[1].vcn - 2);
-	errno = ret;
+	errno = err;
 undo_alloc:
-	ret = errno;
-	if (status.added_cluster) {
-		/* Truncate the last run in the runlist by one cluster. */
-		rl->length--;
-		rl[1].vcn--;
-	} else if (status.added_run) {
-		lcn = rl->lcn;
-		/* Remove the last run from the runlist. */
-		rl->lcn = rl[1].lcn;
-		rl->length = 0;
-	}
-	/* Deallocate the cluster. */
-	if (ntfs_bitmap_clear_bit(lcnbmp_na, lcn))
+	err = errno;
+
+	/* Remove the last run from the runlist. */
+	lcn = rl->lcn;
+	rl->lcn = rl[1].lcn;
+	rl->length = 0;
+	
+	/* FIXME: use an ntfs_cluster_free_* function */
+	if (ntfs_bitmap_clear_bit(vol->lcnbmp_na, lcn))
 		ntfs_log_error("Failed to free cluster.%s\n", es);
-	if (status.mp_rebuilt) {
+	else
+		vol->free_clusters++;
+	if (mp_rebuilt) {
 		if (ntfs_mapping_pairs_build(vol, (u8*)a +
 				le16_to_cpu(a->mapping_pairs_offset),
 				old_alen - le16_to_cpu(a->mapping_pairs_offset),
@@ -782,7 +744,7 @@ undo_alloc:
 	}
 	if (ctx)
 		ntfs_attr_put_search_ctx(ctx);
-	errno = ret;
+	errno = err;
 	return -1;
 }
 
@@ -808,10 +770,9 @@ static int ntfs_mft_bitmap_extend_initialized(ntfs_volume *vol)
 
 	mftbmp_na = vol->mftbmp_na;
 	ctx = ntfs_attr_get_search_ctx(mftbmp_na->ni, NULL);
-	if (!ctx) {
-		ntfs_log_error("Failed to get search context.\n");
+	if (!ctx)
 		return -1;
-	}
+
 	if (ntfs_attr_lookup(mftbmp_na->type, mftbmp_na->name,
 			mftbmp_na->name_len, 0, 0, NULL, 0, ctx)) {
 		ntfs_log_error("Failed to find first attribute extent of "
@@ -844,10 +805,9 @@ static int ntfs_mft_bitmap_extend_initialized(ntfs_volume *vol)
 		err = EIO;
 	/* Try to recover from the error. */
 	ctx = ntfs_attr_get_search_ctx(mftbmp_na->ni, NULL);
-	if (!ctx) {
-		ntfs_log_error("Failed to get search context.%s\n", es);
+	if (!ctx)
 		goto err_out;
-	}
+
 	if (ntfs_attr_lookup(mftbmp_na->type, mftbmp_na->name,
 			mftbmp_na->name_len, 0, 0, NULL, 0, ctx)) {
 		ntfs_log_error("Failed to find first attribute extent of "
@@ -936,7 +896,7 @@ static int ntfs_mft_data_extend_allocation(ntfs_volume *vol)
 		if (rl2)
 			break;
 		if (errno != ENOSPC || nr == min_nr) {
-			ntfs_log_error("Failed to allocate the minimal "
+			ntfs_log_perror("Failed to allocate the minimal "
 					"number of clusters (%lli) for the "
 					"mft data attribute.\n", (long long)nr);
 			return -1;
@@ -969,10 +929,9 @@ static int ntfs_mft_data_extend_allocation(ntfs_volume *vol)
 		;
 	/* Update the attribute record as well. */
 	ctx = ntfs_attr_get_search_ctx(mft_na->ni, NULL);
-	if (!ctx) {
-		ntfs_log_error("Failed to get search context.\n");
+	if (!ctx)
 		goto undo_alloc;
-	}
+
 	if (ntfs_attr_lookup(mft_na->type, mft_na->name, mft_na->name_len, 0,
 			rl[1].vcn, NULL, 0, ctx)) {
 		ntfs_log_error("Failed to find last attribute extent of "
@@ -1007,8 +966,7 @@ static int ntfs_mft_data_extend_allocation(ntfs_volume *vol)
 		// this extent is not required to find the mft record in
 		// question.
 		errno = EOPNOTSUPP;
-		ntfs_log_perror("Not enough space to extended mft data "
-				"attribute.\n");
+		ntfs_log_perror("Not enough space to extended mft data");
 		goto undo_alloc;
 	}
 	mp_rebuilt = TRUE;
@@ -1098,6 +1056,114 @@ undo_alloc:
 	return -1;
 }
 
+
+static int ntfs_mft_record_init(ntfs_volume *vol, s64 size)
+{
+	int ret = -1;
+	ntfs_attr *mft_na, *mftbmp_na;
+	s64 old_data_initialized, old_data_size;
+	ntfs_attr_search_ctx *ctx;
+	
+	ntfs_log_trace("Entering\n");
+	
+	/* NOTE: Caller must sanity check vol, vol->mft_na and vol->mftbmp_na */
+	
+	mft_na = vol->mft_na;
+	mftbmp_na = vol->mftbmp_na;
+	
+	/*
+	 * The mft record is outside the initialized data. Extend the mft data
+	 * attribute until it covers the allocated record. The loop is only
+	 * actually traversed more than once when a freshly formatted volume
+	 * is first written to so it optimizes away nicely in the common case.
+	 */
+	ntfs_log_debug("Status of mft data before extension: "
+			"allocated_size 0x%llx, data_size 0x%llx, "
+			"initialized_size 0x%llx.\n",
+			(long long)mft_na->allocated_size,
+			(long long)mft_na->data_size,
+			(long long)mft_na->initialized_size);
+	while (size > mft_na->allocated_size) {
+		if (ntfs_mft_data_extend_allocation(vol))
+			goto out;
+		ntfs_log_debug("Status of mft data after allocation extension: "
+				"allocated_size 0x%llx, data_size 0x%llx, "
+				"initialized_size 0x%llx.\n",
+				(long long)mft_na->allocated_size,
+				(long long)mft_na->data_size,
+				(long long)mft_na->initialized_size);
+	}
+	
+	old_data_initialized = mft_na->initialized_size;
+	old_data_size = mft_na->data_size;
+	
+	/*
+	 * Extend mft data initialized size (and data size of course) to reach
+	 * the allocated mft record, formatting the mft records along the way.
+	 * Note: We only modify the ntfs_attr structure as that is all that is
+	 * needed by ntfs_mft_record_format().  We will update the attribute
+	 * record itself in one fell swoop later on.
+	 */
+	while (size > mft_na->initialized_size) {
+		s64 ll2 = mft_na->initialized_size >> vol->mft_record_size_bits;
+		mft_na->initialized_size += vol->mft_record_size;
+		if (mft_na->initialized_size > mft_na->data_size)
+			mft_na->data_size = mft_na->initialized_size;
+		ntfs_log_debug("Initializing mft record 0x%llx.\n", (long long)ll2);
+		if (ntfs_mft_record_format(vol, ll2) < 0) {
+			ntfs_log_perror("Failed to format mft record");
+			goto undo_data_init;
+		}
+	}
+	
+	/* Update the mft data attribute record to reflect the new sizes. */
+	ctx = ntfs_attr_get_search_ctx(mft_na->ni, NULL);
+	if (!ctx)
+		goto undo_data_init;
+
+	if (ntfs_attr_lookup(mft_na->type, mft_na->name, mft_na->name_len, 0,
+			0, NULL, 0, ctx)) {
+		ntfs_log_error("Failed to find first attribute extent of "
+				"mft data attribute.\n");
+		ntfs_attr_put_search_ctx(ctx);
+		goto undo_data_init;
+	}
+	ctx->attr->initialized_size = cpu_to_sle64(mft_na->initialized_size);
+	ctx->attr->data_size = cpu_to_sle64(mft_na->data_size);
+	
+	/* Ensure the changes make it to disk. */
+	ntfs_inode_mark_dirty(ctx->ntfs_ino);
+	ntfs_attr_put_search_ctx(ctx);
+	ntfs_log_debug("Status of mft data after mft record initialization: "
+			"allocated_size 0x%llx, data_size 0x%llx, "
+			"initialized_size 0x%llx.\n",
+			(long long)mft_na->allocated_size,
+			(long long)mft_na->data_size,
+			(long long)mft_na->initialized_size);
+	
+	/* Sanity checks. */
+	if (mft_na->data_size > mft_na->allocated_size ||
+	    mft_na->initialized_size > mft_na->data_size)
+		NTFS_BUG("mft_na sanity checks failed");
+	// BUG_ON(mft_na->initialized_size > mft_na->data_size);
+	// BUG_ON(mft_na->data_size > mft_na->allocated_size);
+	
+	/* Sync MFT to minimize data loss if there won't be clean unmount. */
+	if (ntfs_inode_sync(mft_na->ni)) {
+		ntfs_log_error("Failed to sync $MFT.");
+		goto undo_data_init;
+	}
+	
+	ret = 0;
+out:	
+	return ret;
+	
+undo_data_init:
+	mft_na->initialized_size = old_data_initialized;
+	mft_na->data_size = old_data_size;
+	goto out;
+}
+
 /**
  * ntfs_mft_record_alloc - allocate an mft record on an ntfs volume
  * @vol:	volume on which to allocate the mft record
@@ -1183,11 +1249,9 @@ undo_alloc:
  */
 ntfs_inode *ntfs_mft_record_alloc(ntfs_volume *vol, ntfs_inode *base_ni)
 {
-	s64 ll, bit, old_data_initialized, old_data_size;
+	s64 ll, bit;
 	ntfs_attr *mft_na, *mftbmp_na;
-	ntfs_attr_search_ctx *ctx;
 	MFT_RECORD *m;
-	ATTR_RECORD *a;
 	ntfs_inode *ni;
 	int err;
 	u16 seq_no, usn;
@@ -1202,8 +1266,10 @@ ntfs_inode *ntfs_mft_record_alloc(ntfs_volume *vol, ntfs_inode *base_ni)
 		errno = EINVAL;
 		return NULL;
 	}
+	
 	mft_na = vol->mft_na;
 	mftbmp_na = vol->mftbmp_na;
+retry:	
 	bit = ntfs_mft_bitmap_find_free_rec(vol, base_ni);
 	if (bit >= 0) {
 		ntfs_log_debug("Found free record (#1), bit 0x%llx.\n",
@@ -1269,99 +1335,18 @@ ntfs_inode *ntfs_mft_record_alloc(ntfs_volume *vol, ntfs_inode *base_ni)
 	ntfs_log_debug("Found free record (#3), bit 0x%llx.\n", (long long)bit);
 found_free_rec:
 	/* @bit is the found free mft record, allocate it in the mft bitmap. */
-	ntfs_log_debug("At found_free_rec.\n");
 	if (ntfs_bitmap_set_bit(mftbmp_na, bit)) {
 		ntfs_log_error("Failed to allocate bit in mft bitmap.\n");
 		goto err_out;
 	}
 	ntfs_log_debug("Set bit 0x%llx in mft bitmap.\n", (long long)bit);
+	
 	/* The mft bitmap is now uptodate.  Deal with mft data attribute now. */
 	ll = (bit + 1) << vol->mft_record_size_bits;
-	if (ll <= mft_na->initialized_size) {
-		ntfs_log_debug("Allocated mft record already initialized.\n");
-		goto mft_rec_already_initialized;
-	}
-	ntfs_log_debug("Initializing allocated mft record.\n");
-	/*
-	 * The mft record is outside the initialized data.  Extend the mft data
-	 * attribute until it covers the allocated record.  The loop is only
-	 * actually traversed more than once when a freshly formatted volume is
-	 * first written to so it optimizes away nicely in the common case.
-	 */
-	ntfs_log_debug("Status of mft data before extension: "
-			"allocated_size 0x%llx, data_size 0x%llx, "
-			"initialized_size 0x%llx.\n",
-			(long long)mft_na->allocated_size,
-			(long long)mft_na->data_size,
-			(long long)mft_na->initialized_size);
-	while (ll > mft_na->allocated_size) {
-		if (ntfs_mft_data_extend_allocation(vol))
+	if (ll > mft_na->initialized_size)
+		if (ntfs_mft_record_init(vol, ll) < 0)
 			goto undo_mftbmp_alloc;
-		ntfs_log_debug("Status of mft data after allocation extension: "
-				"allocated_size 0x%llx, data_size 0x%llx, "
-				"initialized_size 0x%llx.\n",
-				(long long)mft_na->allocated_size,
-				(long long)mft_na->data_size,
-				(long long)mft_na->initialized_size);
-	}
-	old_data_initialized = mft_na->initialized_size;
-	old_data_size = mft_na->data_size;
-	/*
-	 * Extend mft data initialized size (and data size of course) to reach
-	 * the allocated mft record, formatting the mft records along the way.
-	 * Note: We only modify the ntfs_attr structure as that is all that is
-	 * needed by ntfs_mft_record_format().  We will update the attribute
-	 * record itself in one fell swoop later on.
-	 */
-	while (ll > mft_na->initialized_size) {
-		s64 ll2 = mft_na->initialized_size >> vol->mft_record_size_bits;
-		mft_na->initialized_size += vol->mft_record_size;
-		if (mft_na->initialized_size > mft_na->data_size)
-			mft_na->data_size = mft_na->initialized_size;
-		ntfs_log_debug("Initializing mft record 0x%llx.\n", (long long)ll2);
-		err = ntfs_mft_record_format(vol, ll2);
-		if (err) {
-			ntfs_log_error("Failed to format mft record.\n");
-			goto undo_data_init;
-		}
-	}
-	/* Update the mft data attribute record to reflect the new sizes. */
-	ctx = ntfs_attr_get_search_ctx(mft_na->ni, NULL);
-	if (!ctx) {
-		ntfs_log_error("Failed to get search context.\n");
-		goto undo_data_init;
-	}
-	if (ntfs_attr_lookup(mft_na->type, mft_na->name, mft_na->name_len, 0,
-			0, NULL, 0, ctx)) {
-		ntfs_log_error("Failed to find first attribute extent of "
-				"mft data attribute.\n");
-		ntfs_attr_put_search_ctx(ctx);
-		goto undo_data_init;
-	}
-	a = ctx->attr;
-	a->initialized_size = cpu_to_sle64(mft_na->initialized_size);
-	a->data_size = cpu_to_sle64(mft_na->data_size);
-	/* Ensure the changes make it to disk. */
-	ntfs_inode_mark_dirty(ctx->ntfs_ino);
-	ntfs_attr_put_search_ctx(ctx);
-	ntfs_log_debug("Status of mft data after mft record initialization: "
-			"allocated_size 0x%llx, data_size 0x%llx, "
-			"initialized_size 0x%llx.\n",
-			(long long)mft_na->allocated_size,
-			(long long)mft_na->data_size,
-			(long long)mft_na->initialized_size);
-	/* Sanity checks. */
-	if (mft_na->data_size > mft_na->allocated_size ||
-			mft_na->initialized_size > mft_na->data_size)
-		NTFS_BUG("mft_na sanity checks failed");
-	// BUG_ON(mft_na->initialized_size > mft_na->data_size);
-	// BUG_ON(mft_na->data_size > mft_na->allocated_size);
-	/* Sync MFT to minimize data loss if there won't be clean unmount. */
-	if (ntfs_inode_sync(mft_na->ni)) {
-		ntfs_log_error("Failed to sync $MFT.");
-		goto undo_data_init;
-	}
-mft_rec_already_initialized:
+
 	/*
 	 * We now have allocated and initialized the mft record.  Need to read
 	 * it from disk and re-format it, preserving the sequence number if it
@@ -1373,29 +1358,22 @@ mft_rec_already_initialized:
 		goto undo_mftbmp_alloc;
 	
 	if (ntfs_mft_record_read(vol, bit, m)) {
-		err = errno;
-		ntfs_log_error("Failed to read mft record.\n");
+		ntfs_log_perror("Error reading mft %lld", (long long)bit);
 		free(m);
-		errno = err;
 		goto undo_mftbmp_alloc;
 	}
 	/* Sanity check that the mft record is really not in use. */
 	if (ntfs_is_file_record(m->magic) && (m->flags & MFT_RECORD_IN_USE)) {
-		ntfs_log_error("Mft record 0x%llx was marked unused in "
-				"mft bitmap but is marked used itself.  "
-				"Corrupt filesystem or library bug!  "
-				"Run chkdsk immediately!\n", (long long)bit);
+		ntfs_log_error("Inode %lld is used but it wasn't marked in "
+			       "$MFT bitmap. Fixed.\n", (long long)bit);
 		free(m);
-		errno = EIO;
-		goto undo_mftbmp_alloc;
+		goto retry;
 	}
 	seq_no = m->sequence_number;
 	usn = *(u16*)((u8*)m + le16_to_cpu(m->usa_ofs));
 	if (ntfs_mft_record_layout(vol, bit, m)) {
-		err = errno;
 		ntfs_log_error("Failed to re-format mft record.\n");
 		free(m);
-		errno = err;
 		goto undo_mftbmp_alloc;
 	}
 	if (le16_to_cpu(seq_no))
@@ -1408,10 +1386,8 @@ mft_rec_already_initialized:
 	/* Now need to open an ntfs inode for the mft record. */
 	ni = ntfs_inode_allocate(vol);
 	if (!ni) {
-		err = errno;
 		ntfs_log_error("Failed to allocate buffer for inode.\n");
 		free(m);
-		errno = err;
 		goto undo_mftbmp_alloc;
 	}
 	ni->mft_no = bit;
@@ -1465,9 +1441,7 @@ mft_rec_already_initialized:
 	ntfs_log_debug("Returning opened, allocated %sinode 0x%llx.\n",
 			base_ni ? "extent " : "", (long long)bit);
 	return ni;
-undo_data_init:
-	mft_na->initialized_size = old_data_initialized;
-	mft_na->data_size = old_data_size;
+
 undo_mftbmp_alloc:
 	err = errno;
 	if (ntfs_bitmap_clear_bit(mftbmp_na, bit))
