@@ -76,7 +76,34 @@
  */
 ntfs_volume *ntfs_volume_alloc(void)
 {
-	return calloc(1, sizeof(ntfs_volume));
+	return ntfs_calloc(sizeof(ntfs_volume));
+}
+
+
+static void ntfs_attr_free(ntfs_attr **na)
+{
+	if (na && *na) {
+		ntfs_attr_close(*na);
+		*na = NULL;
+	}
+}
+
+static int ntfs_inode_free(ntfs_inode **ni)
+{
+	int ret = -1;
+
+	if (ni && *ni) {
+		ret = ntfs_inode_close(*ni);
+		*ni = NULL;
+	} 
+	
+	return ret;
+}
+
+static void ntfs_error_set(int *err)
+{
+	if (!*err)
+		*err = errno;
 }
 
 /**
@@ -87,41 +114,51 @@ ntfs_volume *ntfs_volume_alloc(void)
  *
  * Returns:
  */
-static void __ntfs_volume_release(ntfs_volume *v)
+static int __ntfs_volume_release(ntfs_volume *v)
 {
+	int err = 0;
+
+	if (ntfs_inode_free(&v->vol_ni))
+		ntfs_error_set(&err);
+	/* 
+	 * FIXME: Inodes must be synced before closing
+	 * attributes, otherwise unmount could fail.
+	 */
 	if (v->lcnbmp_ni && NInoDirty(v->lcnbmp_ni))
 		ntfs_inode_sync(v->lcnbmp_ni);
-	if (v->vol_ni)
-		ntfs_inode_close(v->vol_ni);
-	if (v->lcnbmp_na)
-		ntfs_attr_close(v->lcnbmp_na);
-	if (v->lcnbmp_ni)
-		ntfs_inode_close(v->lcnbmp_ni);
+	ntfs_attr_free(&v->lcnbmp_na);
+	if (ntfs_inode_free(&v->lcnbmp_ni))
+		ntfs_error_set(&err);
+	
 	if (v->mft_ni && NInoDirty(v->mft_ni))
 		ntfs_inode_sync(v->mft_ni);
-	if (v->mftbmp_na)
-		ntfs_attr_close(v->mftbmp_na);
-	if (v->mft_na)
-		ntfs_attr_close(v->mft_na);
-	if (v->mft_ni)
-		ntfs_inode_close(v->mft_ni);
+	ntfs_attr_free(&v->mftbmp_na);
+	ntfs_attr_free(&v->mft_na);
+	if (ntfs_inode_free(&v->mft_ni))
+		ntfs_error_set(&err);
+	
 	if (v->mftmirr_ni && NInoDirty(v->mftmirr_ni))
 		ntfs_inode_sync(v->mftmirr_ni);
-	if (v->mftmirr_na)
-		ntfs_attr_close(v->mftmirr_na);
-	if (v->mftmirr_ni)
-		ntfs_inode_close(v->mftmirr_ni);
+	ntfs_attr_free(&v->mftmirr_na);
+	if (ntfs_inode_free(&v->mftmirr_ni))
+		ntfs_error_set(&err);
+	
 	if (v->dev) {
 		struct ntfs_device *dev = v->dev;
 
-		dev->d_ops->sync(dev);
+		if (dev->d_ops->sync(dev))
+			ntfs_error_set(&err);
 		if (dev->d_ops->close(dev))
-			ntfs_log_perror("Failed to close the device");
+			ntfs_error_set(&err);
 	}
+
 	free(v->vol_name);
 	free(v->upcase);
 	free(v->attrdef);
 	free(v);
+
+	errno = err;
+	return errno ? -1 : 0;
 }
 
 static void ntfs_attr_setup_flag(ntfs_inode *ni)
@@ -182,10 +219,9 @@ static int ntfs_mft_load(ntfs_volume *vol)
 		goto io_error_exit;
 	}
 	ctx = ntfs_attr_get_search_ctx(vol->mft_ni, NULL);
-	if (!ctx) {
-		ntfs_log_perror("Failed to allocate attribute search context");
+	if (!ctx)
 		goto error_exit;
-	}
+
 	if (p2n(ctx->attr) < p2n(mb) ||
 			(char*)ctx->attr > (char*)mb + vol->mft_record_size) {
 		ntfs_log_error("$MFT is corrupt.\n");
@@ -397,6 +433,7 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 
 	if (!dev || !dev->d_ops || !dev->d_name) {
 		errno = EINVAL;
+		ntfs_log_perror("%s: dev = %p", __FUNCTION__, dev);
 		return NULL;
 	}
 
@@ -408,6 +445,7 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 	vol = ntfs_volume_alloc();
 	if (!vol)
 		goto error_exit;
+	
 	/* Create the default upcase table. */
 	vol->upcase_len = 65536;
 	vol->upcase = ntfs_malloc(vol->upcase_len * sizeof(ntfschar));
@@ -416,18 +454,20 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 	
 	ntfs_upcase_table_build(vol->upcase,
 			vol->upcase_len * sizeof(ntfschar));
+	
 	if (flags & MS_RDONLY)
 		NVolSetReadOnly(vol);
 	if (flags & MS_NOATIME)
 		NVolSetNoATime(vol);
-	ntfs_log_debug("Reading bootsector... ");
-	if (dev->d_ops->open(dev, NVolReadOnly(vol) ? O_RDONLY: O_RDWR)) {
-		ntfs_log_debug(FAILED);
+	
+	/* ...->open needs bracketing to compile with glibc 2.7 */
+	if ((dev->d_ops->open)(dev, NVolReadOnly(vol) ? O_RDONLY: O_RDWR)) {
 		ntfs_log_perror("Error opening partition device");
 		goto error_exit;
 	}
 	/* Attach the device to the volume. */
 	vol->dev = dev;
+	
 	/* Now read the bootsector. */
 	br = ntfs_pread(dev, 0, sizeof(NTFS_BOOT_SECTOR), bs);
 	if (br != sizeof(NTFS_BOOT_SECTOR)) {
@@ -435,13 +475,11 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 		if (br != -1)
 			errno = EINVAL;
 		if (!br)
-			ntfs_log_error("Partition is smaller than bootsector "
-				       "size.\n");
+			ntfs_log_error("Failed to read bootsector (size=0)\n");
 		else
 			ntfs_log_perror("Error reading bootsector");
 		goto error_exit;
 	}
-	ntfs_log_debug(OK);
 	if (!ntfs_boot_sector_is_ntfs(bs)) {
 		errno = EINVAL;
 		goto error_exit;
@@ -513,22 +551,16 @@ ntfs_volume *ntfs_volume_startup(struct ntfs_device *dev, unsigned long flags)
 	 */
 
 	/* Need to setup $MFT so we can use the library read functions. */
-	ntfs_log_debug("Loading $MFT... ");
 	if (ntfs_mft_load(vol) < 0) {
-		ntfs_log_debug(FAILED);
 		ntfs_log_perror("Failed to load $MFT");
 		goto error_exit;
 	}
-	ntfs_log_debug(OK);
 
 	/* Need to setup $MFTMirr so we can use the write functions, too. */
-	ntfs_log_debug("Loading $MFTMirr... ");
 	if (ntfs_mftmirr_load(vol) < 0) {
-		ntfs_log_debug(FAILED);
 		ntfs_log_perror("Failed to load $MFTMirr");
 		goto error_exit;
 	}
-	ntfs_log_debug(OK);
 	return vol;
 error_exit:
 	eo = errno;
@@ -552,23 +584,27 @@ static int ntfs_volume_check_logfile(ntfs_volume *vol)
 	RESTART_PAGE_HEADER *rp = NULL;
 	int err = 0;
 
-	if ((ni = ntfs_inode_open(vol, FILE_LogFile)) == NULL) {
+	ni = ntfs_inode_open(vol, FILE_LogFile);
+	if (!ni) {
 		ntfs_log_perror("Failed to open inode FILE_LogFile");
 		errno = EIO;
 		return -1;
 	}
-	if ((na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0)) == NULL) {
+	
+	na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0);
+	if (!na) {
 		ntfs_log_perror("Failed to open $FILE_LogFile/$DATA");
 		err = EIO;
-		goto exit;
+		goto out;
 	}
+	
 	if (!ntfs_check_logfile(na, &rp) || !ntfs_is_logfile_clean(na, rp))
 		err = EOPNOTSUPP;
 	free(rp);
-exit:
-	if (na)
-		ntfs_attr_close(na);
-	ntfs_inode_close(ni);
+	ntfs_attr_close(na);
+out:	
+	if (ntfs_inode_close(ni))
+		ntfs_error_set(&err);
 	if (err) {
 		errno = err;
 		return -1;
@@ -622,7 +658,10 @@ static ntfs_inode *ntfs_hiberfile_open(ntfs_volume *vol)
 		goto out;
 	}
 out:
-	ntfs_inode_close(ni_root);
+	if (ntfs_inode_close(ni_root)) {
+		ntfs_inode_close(ni_hibr);
+		ni_hibr = NULL;
+	}
 	free(unicode);
 	return ni_hibr;
 }
@@ -642,7 +681,7 @@ static int ntfs_volume_check_hiberfile(ntfs_volume *vol)
 {
 	ntfs_inode *ni;
 	ntfs_attr *na = NULL;
-	int i, bytes_read, ret = -1;
+	int i, bytes_read, err;
 	char *buf = NULL;
 
 	ni = ntfs_hiberfile_open(vol);
@@ -678,21 +717,17 @@ static int ntfs_volume_check_hiberfile(ntfs_volume *vol)
 		errno = EPERM;
 		goto out;
 	}
-	for (i = 0; i < NTFS_HIBERFILE_HEADER_SIZE; i++) {
-		if (buf[i]) {
-			ntfs_log_error("Windows is hibernated, won't mount.\n");
-			errno = EPERM;
-			goto out;
-		}
-	}
         /* All right, all header bytes are zero */
-	ret = 0;
+	errno = 0;
 out:
 	if (na)
 		ntfs_attr_close(na);
 	free(buf);
-	ntfs_inode_close(ni);
-	return ret;
+	err = errno;
+	if (ntfs_inode_close(ni))
+		ntfs_error_set(&err);
+	errno = err;
+	return errno ? -1 : 0;
 }
 
 /**
@@ -737,10 +772,8 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long flags)
 	u32 u;
 
 	vol = ntfs_volume_startup(dev, flags);
-	if (!vol) {
-		ntfs_log_perror("Failed to startup volume");
+	if (!vol)
 		return NULL;
-	}
 
 	/* Load data from $MFT and $MFTMirr and compare the contents. */
 	m  = ntfs_malloc(vol->mftmirr_size << vol->mft_record_size_bits);
@@ -764,20 +797,11 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long flags)
 	l = ntfs_attr_mst_pread(vol->mftmirr_na, 0, vol->mftmirr_size,
 			vol->mft_record_size, m2);
 	if (l != vol->mftmirr_size) {
-		if (l == 4)
-			vol->mftmirr_size = 4;
-		else {
-			if (l == -1)
-				ntfs_log_perror("Failed to read $MFTMirr");
-			else {
-				ntfs_log_error("Failed to read $MFTMirr "
-					       "unexpected length (%d != %lld)."
-					       "\n", vol->mftmirr_size,
-					       (long long)l);
-				errno = EIO;
-			}
+		if (l == -1) {
+			ntfs_log_perror("Failed to read $MFTMirr");
 			goto error_exit;
 		}
+		vol->mftmirr_size = l;
 	}
 	ntfs_log_debug("Comparing $MFTMirr to $MFT... ");
 	for (i = 0; i < vol->mftmirr_size; ++i) {
@@ -908,8 +932,10 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long flags)
 	/* Done with the $UpCase mft record. */
 	ntfs_log_debug(OK);
 	ntfs_attr_close(na);
-	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode, leaking memory");
+	if (ntfs_inode_close(ni)) {
+		ntfs_log_perror("Failed to close $UpCase");
+		goto error_exit;
+	}
 
 	/*
 	 * Now load $Volume and set the version information and flags in the
@@ -924,11 +950,9 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long flags)
 	}
 	/* Get a search context for the $Volume/$VOLUME_INFORMATION lookup. */
 	ctx = ntfs_attr_get_search_ctx(vol->vol_ni, NULL);
-	if (!ctx) {
-		ntfs_log_debug(FAILED);
-		ntfs_log_perror("Failed to allocate attribute search context");
+	if (!ctx)
 		goto error_exit;
-	}
+
 	/* Find the $VOLUME_INFORMATION attribute. */
 	if (ntfs_attr_lookup(AT_VOLUME_INFORMATION, AT_UNNAMED, 0, 0, 0, NULL,
 			0, ctx)) {
@@ -1068,17 +1092,24 @@ ntfs_volume *ntfs_device_mount(struct ntfs_device *dev, unsigned long flags)
 	/* Done with the $AttrDef mft record. */
 	ntfs_log_debug(OK);
 	ntfs_attr_close(na);
-	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode, leaking memory");
+	if (ntfs_inode_close(ni)) {
+		ntfs_log_perror("Failed to close $AttrDef");
+		goto error_exit;
+	}
 	/*
 	 * Check for dirty logfile and hibernated Windows.
 	 * We care only about read-write mounts.
 	 */
 	if (!(flags & MS_RDONLY)) {
-		if (ntfs_volume_check_logfile(vol) < 0)
-			goto error_exit;
 		if (ntfs_volume_check_hiberfile(vol) < 0)
 			goto error_exit;
+		if (ntfs_volume_check_logfile(vol) < 0) {
+			if (!(flags & MS_FORCE))
+				goto error_exit;
+			ntfs_log_info("WARNING: Forced mount, reset $LogFile.\n");
+			if (ntfs_logfile_reset(vol))
+				goto error_exit;
+		}
 	}
 
 	return vol;
@@ -1151,42 +1182,6 @@ ntfs_volume *ntfs_mount(const char *name __attribute__((unused)),
 #endif
 }
 
-/**
- * ntfs_device_umount - close ntfs volume
- * @vol: address of ntfs_volume structure of volume to close
- * @force: if true force close the volume even if it is busy
- *
- * Deallocate all structures (including @vol itself) associated with the ntfs
- * volume @vol.
- *
- * Note it is up to the caller to destroy the device associated with the volume
- * being unmounted after this function returns.
- *
- * Return 0 on success. On error return -1 with errno set appropriately
- * (most likely to one of EAGAIN, EBUSY or EINVAL). The EAGAIN error means that
- * an operation is in progress and if you try the close later the operation
- * might be completed and the close succeed.
- *
- * If @force is true (i.e. not zero) this function will close the volume even
- * if this means that data might be lost.
- *
- * @vol must have previously been returned by a call to ntfs_device_mount().
- *
- * @vol itself is deallocated and should no longer be dereferenced after this
- * function returns success. If it returns an error then nothing has been done
- * so it is safe to continue using @vol.
- */
-int ntfs_device_umount(ntfs_volume *vol,
-		const BOOL force __attribute__((unused)))
-{
-	if (!vol) {
-		errno = EINVAL;
-		return -1;
-	}
-	__ntfs_volume_release(vol);
-	return 0;
-}
-
 /**
  * ntfs_umount - close ntfs volume
  * @vol: address of ntfs_volume structure of volume to close
@@ -1209,19 +1204,19 @@ int ntfs_device_umount(ntfs_volume *vol,
  * function returns success. If it returns an error then nothing has been done
  * so it is safe to continue using @vol.
  */
-int ntfs_umount(ntfs_volume *vol,
-		const BOOL force __attribute__((unused)))
+int ntfs_umount(ntfs_volume *vol, const BOOL force __attribute__((unused)))
 {
 	struct ntfs_device *dev;
+	int ret;
 
 	if (!vol) {
 		errno = EINVAL;
 		return -1;
 	}
 	dev = vol->dev;
-	__ntfs_volume_release(vol);
+	ret = __ntfs_volume_release(vol);
 	ntfs_device_free(dev);
-	return 0;
+	return ret;
 }
 
 #ifdef HAVE_MNTENT_H
@@ -1399,12 +1394,14 @@ int ntfs_logfile_reset(ntfs_volume *vol)
 		return -1;
 	}
 
-	if ((ni = ntfs_inode_open(vol, FILE_LogFile)) == NULL) {
-		ntfs_log_perror("Failed to open inode FILE_LogFile.");
+	ni = ntfs_inode_open(vol, FILE_LogFile);
+	if (!ni) {
+		ntfs_log_perror("Failed to open inode FILE_LogFile");
 		return -1;
 	}
 
-	if ((na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0)) == NULL) {
+	na = ntfs_attr_open(ni, AT_DATA, AT_UNNAMED, 0);
+	if (!na) {
 		eo = errno;
 		ntfs_log_perror("Failed to open $FILE_LogFile/$DATA");
 		goto error_exit;
@@ -1412,10 +1409,10 @@ int ntfs_logfile_reset(ntfs_volume *vol)
 
 	if (ntfs_empty_logfile(na)) {
 		eo = errno;
-		ntfs_log_perror("Failed to empty $FILE_LogFile/$DATA");
 		ntfs_attr_close(na);
 		goto error_exit;
 	}
+	
 	ntfs_attr_close(na);
 	return ntfs_inode_close(ni);
 
@@ -1448,10 +1445,9 @@ int ntfs_volume_write_flags(ntfs_volume *vol, const u16 flags)
 	}
 	/* Get a pointer to the volume information attribute. */
 	ctx = ntfs_attr_get_search_ctx(vol->vol_ni, NULL);
-	if (!ctx) {
-		ntfs_log_perror("Failed to allocate attribute search context");
+	if (!ctx)
 		return -1;
-	}
+
 	if (ntfs_attr_lookup(AT_VOLUME_INFORMATION, AT_UNNAMED, 0, 0, 0, NULL,
 			0, ctx)) {
 		ntfs_log_error("Attribute $VOLUME_INFORMATION was not found "
