@@ -62,6 +62,8 @@
 #endif
 #include <getopt.h>
 #include <syslog.h>
+#include <sys/wait.h>
+#include <pwd.h>
 
 #ifdef HAVE_SETXATTR
 #include <sys/xattr.h>
@@ -104,7 +106,6 @@ typedef enum {
 typedef struct {
 	ntfs_volume *vol;
 	int state;
-	long free_clusters;
 	long free_mft;
 	unsigned int uid;
 	unsigned int gid;
@@ -121,8 +122,6 @@ typedef struct {
 } ntfs_fuse_context_t;
 
 typedef enum {
-	NF_FreeClustersOutdate	= (1 << 0),  /* Information about amount of
-						free clusters is outdated. */
 	NF_FreeMFTOutdate	= (1 << 1),  /* Information about amount of
 						free MFT records is outdated. */
 } ntfs_fuse_state_bits;
@@ -131,8 +130,6 @@ static struct options {
 	char	*mnt_point;	/* Mount point */
 	char	*options;	/* Mount options */
 	char	*device;	/* Device to mount */
-	int	 quiet;		/* Less output */
-	int	 verbose;	/* Extra output */
 } opts;
 
 static const char *EXEC_NAME = "ntfs-3g";
@@ -143,23 +140,32 @@ static u32 ntfs_sequence;
 static const char *locale_msg =
 "WARNING: Couldn't set locale to '%s' thus some file names may not\n"
 "         be correct or visible. Please see the potential solution at\n"
-"         http://www.ntfs-3g.org/support.html#locale\n";
+"         http://ntfs-3g.org/support.html#locale\n";
 
-static const char *fuse26_kmod_msg =
-"WARNING: Deficient Linux kernel detected. Some driver features are\n"
-"         not available (swap file on NTFS, boot from NTFS by LILO), and\n"
-"         unmount is not safe unless it's made sure the ntfs-3g process\n"
-"         naturally terminates after calling 'umount'. If you wish this\n"
-"         message to disappear then you should upgrade to at least kernel\n"
-"         version 2.6.20, or request help from your distribution to fix\n"
-"         the kernel problem. The below web page has more information:\n"
-"         http://www.ntfs-3g.org/support.html#fuse26\n"
-"\n";
+static const char *dev_fuse_msg =
+"HINT: You should be root, or make ntfs-3g setuid root, or load the FUSE\n"
+"      kernel module as root (modprobe fuse) and make sure /dev/fuse exist.\n";
+
+static const char *usage_msg = 
+"\n"
+"%s %s - Third Generation NTFS Driver\n"
+"\n"
+"Copyright (C) 2005-2006 Yura Pakhuchiy\n"
+"Copyright (C) 2006-2007 Szabolcs Szakacsits\n"
+"\n"
+"Usage:    %s <device|image_file> <mount_point> [-o option[,...]]\n"
+"\n"
+"Options:  ro, force, locale=, uid=, gid=, umask=, fmask=, dmask=,\n"
+"          streams_interface=. Please see details in the manual.\n"
+"\n"
+"Example:  ntfs-3g /dev/sda1 /mnt/win -o force,locale=en_EN.UTF-8\n"
+"\n"
+"%s";
 
 static __inline__ void ntfs_fuse_mark_free_space_outdated(void)
 {
-	/* Mark information about free MFT record and clusters outdated. */
-	ctx->state |= (NF_FreeClustersOutdate | NF_FreeMFTOutdate);
+	/* Mark information about free MFT records outdated. */
+	ctx->state |= NF_FreeMFTOutdate;
 }
 
 /**
@@ -212,14 +218,12 @@ static long ntfs_fuse_get_nr_free_mft_records(ntfs_volume *vol, s64 numof_inode)
 	return nr_free;
 }
 
-static long ntfs_fuse_get_nr_free_clusters(ntfs_volume *vol)
+static long ntfs_get_nr_free_clusters(ntfs_volume *vol)
 {
 	u8 *buf;
 	long nr_free = 0;
 	s64 br, total = 0;
 
-	if (!(ctx->state & NF_FreeClustersOutdate))
-		return ctx->free_clusters;
 	buf = ntfs_malloc(vol->cluster_size);
 	if (!buf)
 		return -errno;
@@ -239,8 +243,6 @@ static long ntfs_fuse_get_nr_free_clusters(ntfs_volume *vol)
 	free(buf);
 	if (!total || br < 0)
 		return -errno;
-	ctx->free_clusters = nr_free;
-	ctx->state &= ~(NF_FreeClustersOutdate);
 	return nr_free;
 }
 
@@ -283,7 +285,7 @@ static int ntfs_fuse_statfs(const char *path __attribute__((unused)),
 	sfs->f_blocks = vol->nr_clusters;
 	
 	/* Free data blocks in file system in units of f_bsize. */
-	size = ntfs_fuse_get_nr_free_clusters(vol);
+	size = vol->free_clusters;
 	if (size < 0)
 		size = 0;
 	
@@ -351,6 +353,12 @@ static int ntfs_fuse_parse_path(const char *org_path, char **path,
 	return 0;
 }
 
+static void set_fuse_error(int *err)
+{
+	if (!*err)
+		*err = -errno;
+}
+
 static int ntfs_fuse_getattr(const char *org_path, struct stat *stbuf)
 {
 	int res = 0;
@@ -469,8 +477,8 @@ static int ntfs_fuse_getattr(const char *org_path, struct stat *stbuf)
 	stbuf->st_ctime = ni->last_mft_change_time;
 	stbuf->st_mtime = ni->last_data_change_time;
 exit:
-	if (ni)
-		ntfs_inode_close(ni);
+	if (ni && ntfs_inode_close(ni))
+		set_fuse_error(&res);
 	free(path);
 	if (stream_name_len)
 		free(stream_name);
@@ -545,8 +553,8 @@ exit:
 		free(intx_file);
 	if (na)
 		ntfs_attr_close(na);
-	if (ni)
-		ntfs_inode_close(ni);
+	if (ni && ntfs_inode_close(ni))
+		set_fuse_error(&res);
 	free(path);
 	if (stream_name_len)
 		free(stream_name);
@@ -559,14 +567,17 @@ static int ntfs_fuse_filler(ntfs_fuse_fill_context_t *fill_ctx,
 		const unsigned dt_type __attribute__((unused)))
 {
 	char *filename = NULL;
+	int ret = 0;
 
 	if (name_type == FILE_NAME_DOS)
 		return 0;
+	
 	if (ntfs_ucstombs(name, name_len, &filename, 0) < 0) {
 		ntfs_log_perror("Skipping unrepresentable filename (inode %llu)",
 				(unsigned long long)MREF(mref));
 		return 0;
 	}
+	
 	if (ntfs_fuse_is_named_data_stream(filename)) {
 		ntfs_log_error("Unable to access '%s' (inode %llu) with "
 				"current named streams access interface.\n",
@@ -574,14 +585,16 @@ static int ntfs_fuse_filler(ntfs_fuse_fill_context_t *fill_ctx,
 		free(filename);
 		return 0;
 	}
+	
 	if (MREF(mref) == FILE_root || MREF(mref) >= FILE_first_user ||
 			ctx->show_sys_files) {
 		struct stat st = { .st_ino = MREF(mref) };
 		
-		fill_ctx->filler(fill_ctx->buf, filename, &st, 0);
+		ret = fill_ctx->filler(fill_ctx->buf, filename, &st, 0);
 	}
+	
 	free(filename);
-	return 0;
+	return ret;
 }
 
 static int ntfs_fuse_readdir(const char *path, void *buf,
@@ -603,7 +616,8 @@ static int ntfs_fuse_readdir(const char *path, void *buf,
 	if (ntfs_readdir(ni, &pos, &fill_ctx,
 			(ntfs_filldir_t)ntfs_fuse_filler))
 		err = -errno;
-	ntfs_inode_close(ni);
+	if (ntfs_inode_close(ni))
+		set_fuse_error(&err);
 	return err;
 }
 
@@ -631,7 +645,8 @@ static int ntfs_fuse_open(const char *org_path,
 			ntfs_attr_close(na);
 		} else
 			res = -errno;
-		ntfs_inode_close(ni);
+		if (ntfs_inode_close(ni))
+			set_fuse_error(&res);
 	} else
 		res = -errno;
 	free(path);
@@ -669,8 +684,8 @@ static int ntfs_fuse_read(const char *org_path, char *buf, size_t size,
 	while (size) {
 		res = ntfs_attr_pread(na, offset, size, buf);
 		if (res < (s64)size)
-			ntfs_log_error("ntfs_attr_pread returned less bytes "
-					"than requested.\n");
+			ntfs_log_perror("ntfs_attr_pread partial write (%lld: "
+				"%lld <> %d)", (s64)offset, (s64)size, res);
 		if (res <= 0) {
 			res = -errno;
 			goto exit;
@@ -684,7 +699,7 @@ exit:
 	if (na)
 		ntfs_attr_close(na);
 	if (ni && ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	free(path);
 	if (stream_name_len)
 		free(stream_name);
@@ -719,7 +734,7 @@ static int ntfs_fuse_write(const char *org_path, const char *buf, size_t size,
 	}
 	while (size) {
 		res = ntfs_attr_pwrite(na, offset, size, buf);
-		if (res < (s64)size && errno != ENOSPC)
+		if (res < (s64)size)
 			ntfs_log_perror("ntfs_attr_pwrite partial write (%lld: "
 				"%lld <> %d)", (s64)offset, (s64)size, res);
 		if (res <= 0) {
@@ -736,7 +751,7 @@ exit:
 	if (na)
 		ntfs_attr_close(na);
 	if (ni && ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	free(path);
 	if (stream_name_len)
 		free(stream_name);
@@ -759,24 +774,23 @@ static int ntfs_fuse_truncate(const char *org_path, off_t size)
 		return stream_name_len;
 	vol = ctx->vol;
 	ni = ntfs_pathname_to_inode(vol, NULL, path);
-	if (!ni) {
-		res = -errno;
+	if (!ni)
 		goto exit;
-	}
+
 	na = ntfs_attr_open(ni, AT_DATA, stream_name, stream_name_len);
-	if (!na) {
-		res = -errno;
+	if (!na)
 		goto exit;
-	}
-	res = ntfs_attr_truncate(na, size);
-	// FIXME: check the usage and the importance of the return value
-	if (res)
-		res = -1;
+
+	if (ntfs_attr_truncate(na, size))
+		goto exit;
+	
 	ntfs_fuse_mark_free_space_outdated();
 	ntfs_attr_close(na);
+	errno = 0;
 exit:
+	res = -errno;
 	if (ni && ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	free(path);
 	if (stream_name_len)
 		free(stream_name);
@@ -851,14 +865,15 @@ static int ntfs_fuse_create(const char *org_path, dev_t type, dev_t dev,
 			ni = ntfs_create(dir_ni, uname, uname_len, type);
 			break;
 	}
-	if (ni)
-		ntfs_inode_close(ni);
-	else
+	if (ni) {
+		if (ntfs_inode_close(ni))
+			set_fuse_error(&res);
+	} else
 		res = -errno;
 exit:
 	free(uname);
-	if (dir_ni)
-		ntfs_inode_close(dir_ni);
+	if (dir_ni && ntfs_inode_close(dir_ni))
+		set_fuse_error(&res);
 	if (utarget)
 		free(utarget);
 	free(path);
@@ -891,7 +906,7 @@ static int ntfs_fuse_create_stream(const char *path,
 	if (ntfs_attr_add(ni, AT_DATA, stream_name, stream_name_len, NULL, 0))
 		res = -errno;
 	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	return res;
 }
 
@@ -930,7 +945,13 @@ static int ntfs_fuse_symlink(const char *to, const char *from)
 	return ntfs_fuse_create(from, S_IFLNK, 0, to);
 }
 
-static int ntfs_fuse_link(const char *old_path, const char *new_path)
+/**
+ * NOTE: About the role of mtime: during rename(3), which is currently 
+ * implemented by the help of link() operations, modification time mustn't
+ * be updated, so we NInoSetNoMtimeUpdate() such inodes after they are opened.
+ * This is not very nice itself but it may be eliminated, in time.
+ */
+static int ntfs_fuse_ln(const char *old_path, const char *new_path, int mtime)
 {
 	char *name;
 	ntfschar *uname = NULL;
@@ -951,6 +972,10 @@ static int ntfs_fuse_link(const char *old_path, const char *new_path)
 		res = -errno;
 		goto exit;
 	}
+	
+	if (!mtime)
+		NInoSetNoMtimeUpdate(ni);
+	
 	/* Generate unicode filename. */
 	name = strrchr(path, '/');
 	name++;
@@ -971,15 +996,24 @@ static int ntfs_fuse_link(const char *old_path, const char *new_path)
 	if (ntfs_link(ni, dir_ni, uname, uname_len))
 		res = -errno;
 exit:
-	if (ni)
-		ntfs_inode_close(ni);
+	/* 
+	 * Must close dir_ni first otherwise ntfs_inode_sync_file_name(ni)
+	 * may fail because ni may not be in parent's index on the disk yet.
+	 */
+	if (dir_ni && ntfs_inode_close(dir_ni))
+		set_fuse_error(&res);
+	if (ni && ntfs_inode_close(ni))
+		set_fuse_error(&res);
 	free(uname);
-	if (dir_ni)
-		ntfs_inode_close(dir_ni);
 	free(path);
 	return res;
 }
 
+static int ntfs_fuse_link(const char *old_path, const char *new_path)
+{
+	return ntfs_fuse_ln(old_path, new_path, 1);
+}
+
 static int ntfs_fuse_rm(const char *org_path)
 {
 	char *name;
@@ -1017,11 +1051,11 @@ static int ntfs_fuse_rm(const char *org_path)
 		res = -errno;
 	ni = NULL;
 exit:
-	if (ni)
-		ntfs_inode_close(ni);
+	if (ni && ntfs_inode_close(ni))
+		set_fuse_error(&res);
 	free(uname);
-	if (dir_ni)
-		ntfs_inode_close(dir_ni);
+	if (dir_ni && ntfs_inode_close(dir_ni))
+		set_fuse_error(&res);
 	free(path);
 	return res;
 }
@@ -1040,7 +1074,7 @@ static int ntfs_fuse_rm_stream(const char *path, ntfschar *stream_name,
 		res = -errno;
 
 	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	return res;
 }
 
@@ -1071,16 +1105,16 @@ static int ntfs_fuse_safe_rename(const char *old_path,
 {
 	int ret;
 
-	ntfs_log_trace("Entering");
+	ntfs_log_trace("Entering\n");
 	
-	ret = ntfs_fuse_link(new_path, tmp);
+	ret = ntfs_fuse_ln(new_path, tmp, 0);
 	if (ret)
 		return ret;
 	
 	ret = ntfs_fuse_unlink(new_path);
 	if (!ret) {
 		
-		ret = ntfs_fuse_link(old_path, new_path);
+		ret = ntfs_fuse_ln(old_path, new_path, 0);
 		if (ret)
 			goto restore;
 		
@@ -1094,7 +1128,7 @@ static int ntfs_fuse_safe_rename(const char *old_path,
 	
 	goto cleanup;
 restore:
-	if (ntfs_fuse_link(tmp, new_path)) {
+	if (ntfs_fuse_ln(tmp, new_path, 0)) {
 err:
 		ntfs_log_perror("Rename failed. Existing file '%s' was renamed "
 				"to '%s'", new_path, tmp);
@@ -1111,7 +1145,7 @@ static int ntfs_fuse_rename_existing_dest(const char *old_path, const char *new_
 	char *tmp;
 	const char *ext = ".ntfs-3g-";
 
-	ntfs_log_trace("Entering");
+	ntfs_log_trace("Entering\n");
 	
 	len = strlen(new_path) + strlen(ext) + 10 + 1; /* wc(str(2^32)) + \0 */
 	tmp = ntfs_malloc(len);
@@ -1154,13 +1188,16 @@ static int ntfs_fuse_rename(const char *old_path, const char *new_path)
 			goto out;
 		}
 		
-		ntfs_inode_close(ni);
+		if (ntfs_inode_close(ni)) {
+			set_fuse_error(&ret);
+			goto out;
+		}
 		
 		ret = ntfs_fuse_rename_existing_dest(old_path, new_path);
 		goto out;
 	}
 
-	ret = ntfs_fuse_link(old_path, new_path);
+	ret = ntfs_fuse_ln(old_path, new_path, 0);
 	if (ret)
 		goto out;
 	
@@ -1194,12 +1231,16 @@ static int ntfs_fuse_rmdir(const char *path)
 static int ntfs_fuse_utime(const char *path, struct utimbuf *buf)
 {
 	ntfs_inode *ni;
+	int res = 0;
 
 	if (ntfs_fuse_is_named_data_stream(path))
 		return -EINVAL; /* n/a for named data streams. */
 	ni = ntfs_pathname_to_inode(ctx->vol, NULL, path);
 	if (!ni)
 		return -errno;
+	
+	NInoSetNoParentMtimeUpdate(ni);
+	
 	if (buf) {
 		ni->last_access_time = buf->actime;
 		ni->last_data_change_time = buf->modtime;
@@ -1215,8 +1256,8 @@ static int ntfs_fuse_utime(const char *path, struct utimbuf *buf)
 	NInoFileNameSetDirty(ni);
 	NInoSetDirty(ni);
 	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
-	return 0;
+		set_fuse_error(&res);
+	return res;
 }
 
 static int ntfs_fuse_bmap(const char *path, size_t blocksize, uint64_t *idx)
@@ -1224,7 +1265,8 @@ static int ntfs_fuse_bmap(const char *path, size_t blocksize, uint64_t *idx)
 	ntfs_inode *ni;
 	ntfs_attr *na;
 	LCN lcn;
-	int ret, cl_per_bl = ctx->vol->cluster_size / blocksize;
+	int ret = 0; 
+	int cl_per_bl = ctx->vol->cluster_size / blocksize;
 
 	if (blocksize > ctx->vol->cluster_size)
 		return -EINVAL;
@@ -1255,13 +1297,11 @@ static int ntfs_fuse_bmap(const char *path, size_t blocksize, uint64_t *idx)
 	lcn = ntfs_rl_vcn_to_lcn(na->rl, *idx / cl_per_bl);
 	*idx = (lcn > 0) ? lcn * cl_per_bl + *idx % cl_per_bl : 0;
 	
-	ret = 0;
-	
 close_attr:
 	ntfs_attr_close(na);
 close_inode:
 	if (ntfs_inode_close(ni))
-		ntfs_log_perror("bmap: failed to close inode");
+		set_fuse_error(&ret);
 	return ret;
 }
 
@@ -1328,8 +1368,8 @@ static int ntfs_fuse_listxattr(const char *path, char *list, size_t size)
 exit:
 	if (actx)
 		ntfs_attr_put_search_ctx(actx);
-	ntfs_inode_close(ni);
-	ntfs_log_debug("return %d\n", ret);
+	if (ntfs_inode_close(ni))
+		set_fuse_error(&ret);
 	return ret;
 }
 
@@ -1395,7 +1435,8 @@ static int ntfs_fuse_getxattr_windows(const char *path, const char *name,
 exit:
 	if (actx)
 		ntfs_attr_put_search_ctx(actx);
-	ntfs_inode_close(ni);
+	if (ntfs_inode_close(ni))
+		set_fuse_error(&ret);
 	return ret;
 }
 
@@ -1445,7 +1486,7 @@ exit:
 		ntfs_attr_close(na);
 	free(lename);
 	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	return res;
 }
 
@@ -1505,7 +1546,7 @@ exit:
 		ntfs_attr_close(na);
 	free(lename);
 	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	return res;
 }
 
@@ -1543,7 +1584,7 @@ static int ntfs_fuse_removexattr(const char *path, const char *name)
 exit:
 	free(lename);
 	if (ntfs_inode_close(ni))
-		ntfs_log_perror("Failed to close inode");
+		set_fuse_error(&res);
 	return res;
 }
 
@@ -1556,9 +1597,10 @@ static void ntfs_fuse_destroy(void)
 	
 	if (ctx->vol) {
 		ntfs_log_info("Unmounting %s (%s)\n", opts.device,
-				ctx->vol->vol_name);
+			      ctx->vol->vol_name);
 		if (ntfs_umount(ctx->vol, FALSE))
-			ntfs_log_perror("Failed to unmount volume");
+			ntfs_log_perror("Failed to cleanly unmount volume %s",
+					opts.device);
 	}
 	free(ctx);
 	ctx = NULL;
@@ -1606,9 +1648,9 @@ static int ntfs_fuse_init(void)
 		return -1;
 	
 	*ctx = (ntfs_fuse_context_t) {
-		.state = NF_FreeClustersOutdate | NF_FreeMFTOutdate,
-		.uid = geteuid(),
-		.gid = getegid(),
+		.state = NF_FreeMFTOutdate,
+		.uid = getuid(),
+		.gid = getgid(),
 		.fmask = 0,
 		.dmask = 0,
 		.streams = NF_STREAMS_INTERFACE_NONE,
@@ -1626,6 +1668,8 @@ static ntfs_volume *ntfs_open(const char *device, char *mntpoint, int blkdev)
 		flags |= MS_RDONLY;
 	if (ctx->noatime)
 		flags |= MS_NOATIME;
+	if (ctx->force)
+		flags |= MS_FORCE;
 
 	ctx->vol = utils_mount_volume(device, mntpoint, flags, ctx->force);
 	return ctx->vol;
@@ -1643,14 +1687,16 @@ static char *parse_mount_options(const char *orig_opts)
 	int default_permissions = 0;
 
 	/*
+	 * FIXME: This is not pretty ...
 	 * +7		fsname=
 	 * +1		comma
 	 * +1		null-terminator
 	 * +21          ,blkdev,blksize=65536
 	 * +20          ,default_permissions
+	 * +70          ,user=<max_64_chars>
 	 * +PATH_MAX	resolved realpath() device name
 	 */
-	ret = ntfs_malloc(strlen(def_opts) + strlen(orig_opts) + 64 + PATH_MAX);
+	ret = ntfs_malloc(strlen(def_opts) + strlen(orig_opts) + 256 + PATH_MAX);
 	if (!ret)
 		return NULL;
 	
@@ -1718,7 +1764,8 @@ static char *parse_mount_options(const char *orig_opts)
 			}
 			sscanf(val, "%o", &ctx->fmask);
 			ctx->dmask = ctx->fmask;
-		       	default_permissions = 1;
+			if (ctx->fmask)
+				default_permissions = 1;
 		} else if (!strcmp(opt, "fmask")) {
 			if (!val) {
 				ntfs_log_error("'fmask' option should have "
@@ -1726,7 +1773,8 @@ static char *parse_mount_options(const char *orig_opts)
 				goto err_exit;
 			}
 			sscanf(val, "%o", &ctx->fmask);
-		       	default_permissions = 1;
+			if (ctx->fmask)
+				default_permissions = 1;
 		} else if (!strcmp(opt, "dmask")) {
 			if (!val) {
 				ntfs_log_error("'dmask' option should have "
@@ -1734,7 +1782,8 @@ static char *parse_mount_options(const char *orig_opts)
 				goto err_exit;
 			}
 			sscanf(val, "%o", &ctx->dmask);
-		       	default_permissions = 1;
+			if (ctx->dmask)
+				default_permissions = 1;
 		} else if (!strcmp(opt, "uid")) {
 			if (!val) {
 				ntfs_log_error("'uid' option should have "
@@ -1849,17 +1898,7 @@ err_exit:
 
 static void usage(void)
 {
-	ntfs_log_info("\n%s %s - Third Generation NTFS Driver\n\n",
-			EXEC_NAME, VERSION);
-	ntfs_log_info("Copyright (C) 2005-2006 Yura Pakhuchiy\n");
-	ntfs_log_info("Copyright (C) 2006-2007 Szabolcs Szakacsits\n\n");
-	ntfs_log_info("Usage:    %s device mount_point [-o options]\n\n", 
-		      EXEC_NAME);
-	ntfs_log_info("Options:  ro, force, locale, uid, gid, umask, fmask, "
-		      "dmask, \n\t"
-		      "  show_sys_files, no_def_opts, streams_interface.\n\t"
-		      "  Please see the details in the manual.\n\n");
-	ntfs_log_info("%s\n", ntfs_home);
+	ntfs_log_info(usage_msg, EXEC_NAME, VERSION, EXEC_NAME, ntfs_home);
 }
 
 #ifndef HAVE_REALPATH
@@ -1872,111 +1911,143 @@ static char *realpath(const char *path, char *resolved_path)
 }
 #endif
 
+static int strappend(char **dest, const char *append)
+{
+	char *p;
+	size_t size;
+	
+	if (!dest)
+		return -1;
+	if (!append)
+		return 0;
+	
+	size = strlen(append) + 1;
+	if (*dest)
+		size += strlen(*dest);
+	
+	p = realloc(*dest, size);
+    	if (!p) {
+		ntfs_log_perror("Memory realloction failed");
+		return -1;
+	}
+	
+	if (*dest)
+		strcat(p, append);
+	else
+		strcpy(p, append);
+	*dest = p;
+	
+	return 0;
+}
+
 /**
  * parse_options - Read and validate the programs command line
- *
  * Read the command line, verify the syntax and parse the options.
- * This function is very long, but quite simple.
  *
- * Return:  1 Success
- *	    0 Error, one or more problems
+ * Return:   0 success, -1 error.
  */
 static int parse_options(int argc, char *argv[])
 {
-	int err = 0, help = 0;
-	int c = -1;
+	int c;
 
-	static const char *sopt = "-o:h?qv";
+	static const char *sopt = "-o:hv";
 	static const struct option lopt[] = {
 		{ "options",	 required_argument,	NULL, 'o' },
 		{ "help",	 no_argument,		NULL, 'h' },
-		{ "quiet",	 no_argument,		NULL, 'q' },
 		{ "verbose",	 no_argument,		NULL, 'v' },
 		{ NULL,		 0,			NULL,  0  }
 	};
 
 	opterr = 0; /* We'll handle the errors, thank you. */
 
-	opts.mnt_point = NULL;
-	opts.options = NULL;
-	opts.device = NULL;
-
 	while ((c = getopt_long(argc, argv, sopt, lopt, NULL)) != -1) {
 		switch (c) {
 		case 1:	/* A non-option argument */
 			if (!opts.device) {
 				opts.device = ntfs_malloc(PATH_MAX + 1);
-				if (!opts.device) {
-					err++;
-					break;
-				}
+				if (!opts.device)
+					return -1;
+				
 				/* We don't want relative path in /etc/mtab. */
 				if (optarg[0] != '/') {
 					if (!realpath(optarg, opts.device)) {
-						ntfs_log_perror("Cannot mount "
-								"'%s'", optarg);
+						ntfs_log_perror("%s: "
+							"Cannot mount '%s'", 
+							EXEC_NAME, optarg);
 						free(opts.device);
 						opts.device = NULL;
-						err++;
-						break;
+						return -1;
 					}
 				} else
 					strcpy(opts.device, optarg);
-			} else if (!opts.mnt_point)
+			} else if (!opts.mnt_point) {
 				opts.mnt_point = optarg;
-			else {
-				ntfs_log_error("You must specify exactly one "
+			} else {
+				ntfs_log_error("%s: You must specify exactly one "
 						"device and exactly one mount "
-						"point.\n");
-				err++;
+						"point.\n", EXEC_NAME);
+				return -1;
 			}
 			break;
 		case 'o':
-			if (!opts.options)
-				opts.options = optarg;
-			else {
-				ntfs_log_error("You must specify exactly one "
-						"set of options.\n");
-				err++;
-			}
+			if (opts.options)
+				if (strappend(&opts.options, ","))
+					return -1;
+			if (strappend(&opts.options, optarg))
+				return -1;
 			break;
 		case 'h':
-		case '?':
-			help++;
-			break;
-		case 'q':
-			opts.quiet++;
-			break;
+			usage();
+			exit(9);
 		case 'v':
-			opts.verbose++;
+			/*
+			 * We must handle the 'verbose' option even if
+			 * we don't use it because mount(8) passes it.
+			 */
 			break;
 		default:
-			ntfs_log_error("Unknown option '%s'.\n",
-					argv[optind - 1]);
-			err++;
-			break;
+			ntfs_log_error("%s: Unknown option '%s'.\n", EXEC_NAME,
+				       argv[optind - 1]);
+			return -1;
 		}
 	}
 
-	if (help) {
-		opts.quiet = 0;
-	} else {
-		if (!opts.device) {
-			ntfs_log_error("No device specified.\n");
-			err++;
-		}
-
-		if (opts.quiet && opts.verbose) {
-			ntfs_log_error("You may not use --quiet and --verbose "
-					"at the same time.\n");
-			err++;
-		}
+	if (!opts.device) {
+		ntfs_log_error("%s: No device is specified.\n", EXEC_NAME);
+		return -1;
+	}
+	if (!opts.mnt_point) {
+		ntfs_log_error("%s: No mountpoint is specified.\n", EXEC_NAME);
+		return -1;
 	}
 
-	if (help || err)
-		usage();
+	return 0;
+}
 
-	return (!help && !err);
+#ifdef linux
+
+static const char *fuse26_kmod_msg =
+"WARNING: Deficient Linux kernel detected. Some driver features are\n"
+"         not available (swap file on NTFS, boot from NTFS by LILO), and\n"
+"         unmount is not safe unless it's made sure the ntfs-3g process\n"
+"         naturally terminates after calling 'umount'. If you wish this\n"
+"         message to disappear then you should upgrade to at least kernel\n"
+"         version 2.6.20, or request help from your distribution to fix\n"
+"         the kernel problem. The below web page has more information:\n"
+"         http://ntfs-3g.org/support.html#fuse26\n"
+"\n";
+
+static void create_dev_fuse(void)
+{
+	struct stat st;
+	
+	if (stat("/dev/fuse", &st) && (errno == ENOENT)) {
+		if (mknod("/dev/fuse", S_IFCHR | 0666, makedev(10, 229))) {
+			ntfs_log_perror("Failed to create /dev/fuse");
+			if (errno == EPERM)
+				ntfs_log_error(dev_fuse_msg);
+		}
+	}
 }
 
 static fuse_fstype get_fuse_fstype(void)
@@ -2003,29 +2074,23 @@ static fuse_fstype get_fuse_fstype(void)
 	return fstype;
 }
 
-static void create_dev_fuse(void)
-{
-	struct stat st;
-	
-	if (stat("/dev/fuse", &st) && (errno == ENOENT)) {
-		if (mknod("/dev/fuse", S_IFCHR | 0666, makedev(10, 229)))
-			ntfs_log_perror("Failed to create /dev/fuse");
-	}
-}
-
 static fuse_fstype load_fuse_module(void)
 {
 	int i;
 	struct stat st;
-	const char *load_fuse_cmd = "/sbin/modprobe fuse";
+	pid_t pid;
+	const char *cmd = "/sbin/modprobe";
 	struct timespec req = { 0, 100000000 };   /* 100 msec */
 	fuse_fstype fstype;
 	
-	if (stat("/sbin/modprobe", &st) == -1)
-		load_fuse_cmd = "modprobe fuse";
-	
-	if (getuid() == 0)
-		system(load_fuse_cmd);
+	if (!stat(cmd, &st) && !getuid()) {
+		pid = fork();
+		if (!pid) {
+			execl(cmd, cmd, "fuse", NULL);
+			_exit(1);
+		} else if (pid != -1)
+			waitpid(pid, NULL, 0);
+	}
 	
 	for (i = 0; i < 10; i++) {
 		/* 
@@ -2042,13 +2107,15 @@ static fuse_fstype load_fuse_module(void)
 	return fstype;
 }
 
+#endif
+
 static struct fuse_chan *try_fuse_mount(char *parsed_options)
 {
 	struct fuse_chan *fc = NULL;
 	struct fuse_args margs = FUSE_ARGS_INIT(0, NULL);
 	
 	/* The fuse_mount() options get modified, so we always rebuild it */
-	if ((fuse_opt_add_arg(&margs, "") == -1 ||
+	if ((fuse_opt_add_arg(&margs, EXEC_NAME) == -1 ||
 	     fuse_opt_add_arg(&margs, "-o") == -1 ||
 	     fuse_opt_add_arg(&margs, parsed_options) == -1)) {
 		ntfs_log_error("Failed to set FUSE options.\n");
@@ -2082,13 +2149,34 @@ static void set_fuseblk_options(char *parsed_options)
 	strcat(parsed_options, options);
 }
 
+static void set_user_mount_option(char *parsed_options, uid_t uid)
+{
+	struct passwd *pw;
+	char option[64];
+	
+	if (!uid)
+		return;
+	
+	errno = 0;
+	pw = getpwuid(uid);
+	if (!pw || !pw->pw_name) {
+		ntfs_log_perror("WARNING: could not get username for uid %lld, "
+				"unprivileged unmount may fail", (long long)uid);
+		return;
+	}
+	
+	/* parsed_options already allocated enough space. */
+	snprintf(option, sizeof(option), ",user=%s", pw->pw_name);
+	strcat(parsed_options, option);
+}
+
 int main(int argc, char *argv[])
 {
 	char *parsed_options = NULL;
 	struct fuse_args margs = FUSE_ARGS_INIT(0, NULL);
 	struct fuse *fh;
 	struct fuse_chan *fc;
-	fuse_fstype fstype;
+	fuse_fstype fstype = FSTYPE_UNKNOWN;
 	struct stat sbuf;
 	int use_blkdev = 0;
 	uid_t uid, euid;
@@ -2099,8 +2187,10 @@ int main(int argc, char *argv[])
 	signal(SIGINT, signal_handler);
 	signal(SIGTERM, signal_handler);
 
-	if (!parse_options(argc, argv))
+	if (parse_options(argc, argv)) {
+		usage();
 		return 1;
+	}
 
 	if (ntfs_fuse_init())
 		return 2;
@@ -2117,11 +2207,13 @@ int main(int argc, char *argv[])
 		goto err_out;
 	}
 
+#ifdef linux
 	fstype = get_fuse_fstype();
 	if (fstype == FSTYPE_NONE || fstype == FSTYPE_UNKNOWN)
 		fstype = load_fuse_module();
 	
 	create_dev_fuse();
+#endif	
 	
 	if (stat(opts.device, &sbuf)) {
 		ntfs_log_perror("Failed to access '%s'", opts.device);
@@ -2134,8 +2226,16 @@ int main(int argc, char *argv[])
 	if (!ntfs_open(opts.device, opts.mnt_point, use_blkdev))
 		goto err_out;
 	
-	if (use_blkdev)
+	ctx->vol->free_clusters = ntfs_get_nr_free_clusters(ctx->vol);
+	if (ctx->vol->free_clusters < 0) {
+		ntfs_log_perror("Failed to read NTFS $Bitmap");
+		goto err_out;
+	}
+	
+	if (use_blkdev) {
 	    set_fuseblk_options(parsed_options);
+	    set_user_mount_option(parsed_options, uid);
+	}
 	
 	/* Libfuse can't always find fusermount, so let's help it. */
 	if (setenv("PATH", ":/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin", 0))
@@ -2153,7 +2253,7 @@ int main(int argc, char *argv[])
 		if (fuse_opt_add_arg(&margs, "use_ino,kernel_cache") == -1)
 			fh = NULL;
 	} else {
-		if (fuse_opt_add_arg(&margs, "use_ino,debug") == -1)
+		if (fuse_opt_add_arg(&margs, "use_ino,kernel_cache,debug") == -1)
 			fh = NULL;
 	}
 	if (fh)
@@ -2172,9 +2272,10 @@ int main(int argc, char *argv[])
 		goto err_out;
 	}
 
+#ifdef linux
 	if (S_ISBLK(sbuf.st_mode) && (fstype == FSTYPE_FUSE))
 		ntfs_log_info(fuse26_kmod_msg);
-	
+#endif	
 	if (!ctx->no_detach) {
 		if (daemon(0, ctx->debug))
 			ntfs_log_error("Failed to daemonize.\n");
@@ -2192,7 +2293,8 @@ int main(int argc, char *argv[])
 			opts.device, (ctx->ro) ? "Read-Only" : "Read-Write",
 			ctx->vol->vol_name, ctx->vol->major_ver,
 			ctx->vol->minor_ver);
-	ntfs_log_info("Options: %s\n", parsed_options);
+	ntfs_log_info("Cmdline options: %s\n", opts.options);
+	ntfs_log_info("Mount options: %s\n", parsed_options);
 	
 	fuse_loop(fh);
 	
@@ -2200,6 +2302,7 @@ int main(int argc, char *argv[])
 	fuse_destroy(fh);
 	err = 0;
 err_out:
+	free(opts.options);
 	free(parsed_options);
 	ntfs_fuse_destroy();
 	return err;
